from enum import Enum
from typing import Union, cast, Final
from pyteal.ast.abi.type import BaseType, ComputedValue, TypeSpec
from pyteal.ast.expr import Expr
from pyteal.ast.int import Int
from pyteal.ast.txn import TxnObject, TxnType
from pyteal.ast.gtxn import Gtxn
from pyteal.types import TealType
from pyteal.errors import TealInputError, TealInternalError


class TransactionType(Enum):
    Any = "txn"
    Payment = "pay"
    KeyRegistration = "keyreg"
    AssetConfig = "acfg"
    AssetTransfer = "axfer"
    AssetFreeze = "afrz"
    ApplicationCall = "appl"


TransactionType.__module__ = "pyteal.abi"


class TransactionTypeSpec(TypeSpec):
    def __init__(self) -> None:
        super().__init__()

    def new_instance(self) -> "Transaction":
        return Transaction()

    def annotation_type(self) -> "type[Transaction]":
        return Transaction

    def is_dynamic(self) -> bool:
        return False

    def byte_length_static(self) -> int:
        raise TealInputError("Transaction Types don't have a static size")

    def storage_type(self) -> TealType:
        return TealType.uint64

    def txn_type_enum(self) -> Expr:
        """Get the integer transaction type value this TransactionTypeSpec represents.

        See :any:`TxnType` for the complete list.

        If this is a generic TransactionTypeSpec, i.e. type :code:`txn`, this method will raise an error, since this type does not represent a single transaction type.
        """
        raise TealInternalError(
            "abi.TransactionTypeSpec does not represent a specific transaction type"
        )

    def __eq__(self, other: object) -> bool:
        return type(self) is type(other)

    def __str__(self) -> str:
        return TransactionType.Any.value


TransactionTypeSpec.__module__ = "pyteal.abi"


class Transaction(BaseType):
    def __init__(self, spec: TransactionTypeSpec = None) -> None:
        if spec is None:
            super().__init__(TransactionTypeSpec())
        else:
            super().__init__(spec)

    def type_spec(self) -> TransactionTypeSpec:
        return cast(TransactionTypeSpec, super().type_spec())

    def get(self) -> TxnObject:
        return Gtxn[self.index()]

    def _set_index(
        self, value: Union[int, Expr, "Transaction", ComputedValue["Transaction"]]
    ) -> Expr:
        match value:
            case ComputedValue():
                return self._set_with_computed_type(value)
            case BaseType():
                return self.stored_value.store(self.stored_value.load())
            case int():
                return self.stored_value.store(Int(value))
            case Expr():
                return self.stored_value.store(value)
            case _:
                raise TealInputError(f"Cant store a {type(value)} in a Transaction")

    def index(self) -> Expr:
        return self.stored_value.load()

    def decode(
        self,
        encoded: Expr,
        *,
        start_index: Expr = None,
        end_index: Expr = None,
        length: Expr = None,
    ) -> Expr:
        raise TealInputError("A Transaction cannot be decoded")

    def encode(self) -> Expr:
        raise TealInputError("A Transaction cannot be encoded")


Transaction.__module__ = "pyteal.abi"


class PaymentTransactionTypeSpec(TransactionTypeSpec):
    def new_instance(self) -> "PaymentTransaction":
        return PaymentTransaction()

    def annotation_type(self) -> "type[PaymentTransaction]":
        return PaymentTransaction

    def txn_type_enum(self) -> Expr:
        return TxnType.Payment

    def __str__(self) -> str:
        return TransactionType.Payment.value


PaymentTransactionTypeSpec.__module__ = "pyteal.abi"


class PaymentTransaction(Transaction):
    def __init__(self):
        super().__init__(PaymentTransactionTypeSpec())


PaymentTransaction.__module__ = "pyteal.abi"


class KeyRegisterTransactionTypeSpec(TransactionTypeSpec):
    def new_instance(self) -> "KeyRegisterTransaction":
        return KeyRegisterTransaction()

    def annotation_type(self) -> "type[KeyRegisterTransaction]":
        return KeyRegisterTransaction

    def txn_type_enum(self) -> Expr:
        return TxnType.KeyRegistration

    def __str__(self) -> str:
        return TransactionType.KeyRegistration.value


KeyRegisterTransactionTypeSpec.__module__ = "pyteal.abi"


class KeyRegisterTransaction(Transaction):
    def __init__(self):
        super().__init__(KeyRegisterTransactionTypeSpec())


KeyRegisterTransaction.__module__ = "pyteal.abi"


class AssetConfigTransactionTypeSpec(TransactionTypeSpec):
    def new_instance(self) -> "AssetConfigTransaction":
        return AssetConfigTransaction()

    def annotation_type(self) -> "type[AssetConfigTransaction]":
        return AssetConfigTransaction

    def txn_type_enum(self) -> Expr:
        return TxnType.AssetConfig

    def __str__(self) -> str:
        return TransactionType.AssetConfig.value


AssetConfigTransactionTypeSpec.__module__ = "pyteal.abi"


class AssetConfigTransaction(Transaction):
    def __init__(self):
        super().__init__(AssetConfigTransactionTypeSpec())


AssetConfigTransaction.__module__ = "pyteal.abi"


class AssetFreezeTransactionTypeSpec(TransactionTypeSpec):
    def new_instance(self) -> "AssetFreezeTransaction":
        return AssetFreezeTransaction()

    def annotation_type(self) -> "type[AssetFreezeTransaction]":
        return AssetFreezeTransaction

    def txn_type_enum(self) -> Expr:
        return TxnType.AssetFreeze

    def __str__(self) -> str:
        return TransactionType.AssetFreeze.value


AssetFreezeTransactionTypeSpec.__module__ = "pyteal.abi"


class AssetFreezeTransaction(Transaction):
    def __init__(self):
        super().__init__(AssetFreezeTransactionTypeSpec())


AssetFreezeTransaction.__module__ = "pyteal.abi"


class AssetTransferTransactionTypeSpec(TransactionTypeSpec):
    def new_instance(self) -> "AssetTransferTransaction":
        return AssetTransferTransaction()

    def annotation_type(self) -> "type[AssetTransferTransaction]":
        return AssetTransferTransaction

    def txn_type_enum(self) -> Expr:
        return TxnType.AssetTransfer

    def __str__(self) -> str:
        return TransactionType.AssetTransfer.value


AssetTransferTransactionTypeSpec.__module__ = "pyteal.abi"


class AssetTransferTransaction(Transaction):
    def __init__(self):
        super().__init__(AssetTransferTransactionTypeSpec())


AssetTransferTransaction.__module__ = "pyteal.abi"


class ApplicationCallTransactionTypeSpec(TransactionTypeSpec):
    def new_instance(self) -> "ApplicationCallTransaction":
        return ApplicationCallTransaction()

    def annotation_type(self) -> "type[ApplicationCallTransaction]":
        return ApplicationCallTransaction

    def txn_type_enum(self) -> Expr:
        return TxnType.ApplicationCall

    def __str__(self) -> str:
        return TransactionType.ApplicationCall.value


ApplicationCallTransactionTypeSpec.__module__ = "pyteal.abi"


class ApplicationCallTransaction(Transaction):
    def __init__(self):
        super().__init__(ApplicationCallTransactionTypeSpec())


ApplicationCallTransaction.__module__ = "pyteal.abi"

TransactionTypeSpecs: Final[list[TypeSpec]] = [
    TransactionTypeSpec(),
    PaymentTransactionTypeSpec(),
    KeyRegisterTransactionTypeSpec(),
    AssetConfigTransactionTypeSpec(),
    AssetFreezeTransactionTypeSpec(),
    AssetTransferTransactionTypeSpec(),
    ApplicationCallTransactionTypeSpec(),
]



from typing import Set
from pyteal.ast import ScratchSlot
from pyteal.ir import TealBlock, TealOp, Op
from pyteal.errors import TealInternalError


class OptimizeOptions:
    """An object which specifies the optimizations to be performed and relevant context.

    _skip_slots: the slots that should be skipped during optimization. At the moment this includes:
            1. reserved slots because they may have dependencies outside
            the current application. For example, the 'gloads' opcode can
            access the slots of other applications in the tx group.
            2. global slots because they're outside the scope of global
            optimizations, which only apply to the control flow graph of
            a single subroutine.
            3. slots used with dynamic scratch vars. These slots use
            indirection by means of the 'stores' opcode and dependencies
            can only be determined at runtime.

    Args:

        scratch_slots (optional): cancel contiguous store/load operations
            that have no load dependencies elsewhere.
    """

    def __init__(self, *, scratch_slots: bool = False):
        self.scratch_slots = scratch_slots
        self._skip_slots: Set[ScratchSlot] = set()


def _remove_extraneous_slot_access(start: TealBlock, remove: Set[ScratchSlot]):
    def keep_op(op: TealOp) -> bool:
        if type(op) != TealOp or (op.op != Op.store and op.op != Op.load):
            return True

        return not set(op.getSlots()).issubset(remove)

    for block in TealBlock.Iterate(start):
        block.ops = list(filter(keep_op, block.ops))


# Very dumb, overly eager dependency checking. A "dependency" is considered
# any time the slot is loaded from in the entire control flow graph. This
# can definitely be improved in the future.
def _has_load_dependencies(
    cur_block: TealBlock, start: TealBlock, slot: ScratchSlot, pos: int
) -> bool:
    for block in TealBlock.Iterate(start):
        for i, op in enumerate(block.ops):
            if block == cur_block and i == pos:
                continue

            if type(op) == TealOp and op.op == Op.load and slot in set(op.getSlots()):
                return True

    return False


def _apply_slot_to_stack(
    cur_block: TealBlock, start: TealBlock, skip_slots: Set[ScratchSlot]
):
    slots_to_remove = set()
    # surprisingly, this slicing is totally safe - even if the list is empty.
    for i, op in enumerate(cur_block.ops[:-1]):
        if type(op) != TealOp or op.op != Op.store:
            continue

        if set(op.getSlots()).issubset(skip_slots):
            continue

        next_op = cur_block.ops[i + 1]
        if type(next_op) != TealOp or next_op.op != Op.load:
            continue

        cur_slots, next_slots = op.getSlots(), next_op.getSlots()
        if len(cur_slots) != 1 or len(next_slots) != 1:
            raise TealInternalError(
                "load/store op does not have exactly one slot argument"
            )
        if cur_slots[0] != next_slots[0]:
            continue

        if not _has_load_dependencies(cur_block, start, cur_slots[0], i + 1):
            slots_to_remove.add(cur_slots[0])

    _remove_extraneous_slot_access(start, slots_to_remove)


def apply_global_optimizations(start: TealBlock, options: OptimizeOptions) -> TealBlock:
    # limit number of iterations to length of teal program to avoid potential
    # infinite loops.
    for block in TealBlock.Iterate(start):
        for _ in range(len(block.ops)):
            prev_ops = block.ops.copy()
            if options.scratch_slots:
                _apply_slot_to_stack(block, start, options._skip_slots)

            if prev_ops == block.ops:
                break

    return start


OptimizeOptions.__module__ = "pyteal"



import pytest

import pyteal as pt

from tests.compile_asserts import assert_new_v_old

# TODO: remove these skips when the following is fixed: https://github.com/algorand/pyteal/issues/199
STABLE_SLOT_GENERATION = False

# ### TESTS FOR NEW PyTEAL THAT USES PASS-BY-REF / DYNAMIC
@pt.Subroutine(pt.TealType.none)
def logcat_dynamic(first: pt.ScratchVar, an_int):
    return pt.Seq(
        first.store(pt.Concat(first.load(), pt.Itob(an_int))),
        pt.Log(first.load()),
    )


def sub_logcat_dynamic():
    first = pt.ScratchVar(pt.TealType.bytes)
    return pt.Seq(
        first.store(pt.Bytes("hello")),
        logcat_dynamic(first, pt.Int(42)),
        pt.Assert(pt.Bytes("hello42") == first.load()),
        pt.Int(1),
    )


def wilt_the_stilt():
    player_score = pt.DynamicScratchVar(pt.TealType.uint64)

    wilt = pt.ScratchVar(pt.TealType.uint64, 129)
    kobe = pt.ScratchVar(pt.TealType.uint64)
    dt = pt.ScratchVar(pt.TealType.uint64, 131)

    return pt.Seq(
        player_score.set_index(wilt),
        player_score.store(pt.Int(100)),
        player_score.set_index(kobe),
        player_score.store(pt.Int(81)),
        player_score.set_index(dt),
        player_score.store(pt.Int(73)),
        pt.Assert(player_score.load() == pt.Int(73)),
        pt.Assert(player_score.index() == pt.Int(131)),
        player_score.set_index(wilt),
        pt.Assert(player_score.load() == pt.Int(100)),
        pt.Assert(player_score.index() == pt.Int(129)),
        pt.Int(100),
    )


@pt.Subroutine(pt.TealType.none)
def swap(x: pt.ScratchVar, y: pt.ScratchVar):
    z = pt.ScratchVar(pt.TealType.anytype)
    return pt.Seq(
        z.store(x.load()),
        x.store(y.load()),
        y.store(z.load()),
    )


@pt.Subroutine(pt.TealType.none)
def cat(x, y):
    return pt.Pop(pt.Concat(x, y))


def swapper():
    a = pt.ScratchVar(pt.TealType.bytes)
    b = pt.ScratchVar(pt.TealType.bytes)
    return pt.Seq(
        a.store(pt.Bytes("hello")),
        b.store(pt.Bytes("goodbye")),
        cat(a.load(), b.load()),
        swap(a, b),
        pt.Assert(a.load() == pt.Bytes("goodbye")),
        pt.Assert(b.load() == pt.Bytes("hello")),
        pt.Int(1000),
    )


@pt.Subroutine(pt.TealType.uint64)
def mixed_annotations(x: pt.Expr, y: pt.Expr, z: pt.ScratchVar) -> pt.Expr:
    return pt.Seq(
        z.store(x),
        pt.Log(pt.Concat(y, pt.Bytes("="), pt.Itob(x))),
        x,
    )


def sub_mixed():
    x = pt.Int(42)
    y = pt.Bytes("x")
    z = pt.ScratchVar(pt.TealType.uint64)
    return mixed_annotations(x, y, z)


def lots_o_vars():
    z = pt.Int(0)
    one = pt.ScratchVar()
    two = pt.ScratchVar()
    three = pt.ScratchVar()
    four = pt.ScratchVar()
    five = pt.Bytes("five")
    six = pt.Bytes("six")
    seven = pt.Bytes("seven")
    eight = pt.Bytes("eight")
    nine = pt.Bytes("nine")
    ten = pt.Bytes("ten")
    eleven = pt.Bytes("eleven")
    twelve = pt.Bytes("twelve")
    int_cursor = pt.DynamicScratchVar(pt.TealType.uint64)
    bytes_cursor = pt.DynamicScratchVar(pt.TealType.bytes)
    thirteen = pt.ScratchVar(pt.TealType.uint64, 13)
    fourteen = pt.ScratchVar(pt.TealType.bytes, 14)
    fifteen = pt.ScratchVar(pt.TealType.uint64)
    sixteen = pt.ScratchVar(pt.TealType.bytes)
    leet = pt.Int(1337)
    ngl = pt.Bytes("NGL: ")
    return (
        pt.If(
            pt.Or(
                pt.App.id() == pt.Int(0), pt.Txn.application_args.length() == pt.Int(0)
            )
        )
        .Then(pt.Int(1))
        .Else(
            pt.Seq(
                one.store(pt.Int(1)),
                two.store(pt.Bytes("two")),
                three.store(pt.Int(3)),
                four.store(pt.Bytes("four")),
                pt.App.localPut(z, five, pt.Int(5)),
                pt.App.localPut(z, six, six),
                pt.App.localPut(z, seven, pt.Int(7)),
                pt.App.localPut(z, eight, eight),
                pt.App.globalPut(nine, pt.Int(9)),
                pt.App.globalPut(ten, ten),
                pt.App.globalPut(eleven, pt.Int(11)),
                pt.App.globalPut(twelve, twelve),
                one.store(one.load() + leet),
                two.store(pt.Concat(ngl, two.load())),
                three.store(three.load() + leet),
                four.store(pt.Concat(ngl, four.load())),
                pt.App.localPut(z, five, leet + pt.App.localGet(z, five)),
                pt.App.localPut(z, six, pt.Concat(ngl, pt.App.localGet(z, six))),
                pt.App.localPut(z, seven, pt.App.localGet(z, seven)),
                pt.App.localPut(z, eight, pt.Concat(ngl, pt.App.localGet(z, eight))),
                pt.App.globalPut(nine, leet + pt.App.globalGet(nine)),
                pt.App.globalPut(ten, pt.Concat(ngl, pt.App.globalGet(ten))),
                pt.App.globalPut(eleven, leet + pt.App.globalGet(eleven)),
                pt.App.globalPut(twelve, pt.Concat(ngl, pt.App.globalGet(twelve))),
                thirteen.store(pt.Btoi(pt.Txn.application_args[0])),
                fourteen.store(pt.Txn.application_args[1]),
                fifteen.store(pt.Btoi(pt.Txn.application_args[2])),
                sixteen.store(pt.Txn.application_args[3]),
                pt.Pop(one.load()),
                pt.Pop(two.load()),
                pt.Pop(three.load()),
                pt.Pop(four.load()),
                pt.Pop(pt.App.localGet(z, five)),
                pt.Pop(pt.App.localGet(z, six)),
                pt.Pop(pt.App.localGet(z, seven)),
                pt.Pop(pt.App.localGet(z, eight)),
                pt.Pop(pt.App.globalGet(nine)),
                pt.Pop(pt.App.globalGet(ten)),
                pt.Pop(pt.App.globalGet(eleven)),
                pt.Pop(pt.App.globalGet(twelve)),
                int_cursor.set_index(thirteen),
                pt.Log(pt.Itob(int_cursor.load())),
                bytes_cursor.set_index(fourteen),
                pt.Log(bytes_cursor.load()),
                int_cursor.set_index(fifteen),
                pt.Log(pt.Itob(int_cursor.load())),
                bytes_cursor.set_index(sixteen),
                pt.Log(bytes_cursor.load()),
                leet,
            )
        )
    )


def empty_scratches():
    cursor = pt.DynamicScratchVar()
    i1 = pt.ScratchVar(pt.TealType.uint64, 0)
    i2 = pt.ScratchVar(pt.TealType.uint64, 2)
    i3 = pt.ScratchVar(pt.TealType.uint64, 4)
    s1 = pt.ScratchVar(pt.TealType.bytes, 1)
    s2 = pt.ScratchVar(pt.TealType.bytes, 3)
    s3 = pt.ScratchVar(pt.TealType.bytes, 5)
    return pt.Seq(
        cursor.set_index(i1),
        cursor.store(pt.Int(0)),
        cursor.set_index(s1),
        cursor.store(pt.Bytes("")),
        cursor.set_index(i2),
        cursor.store(pt.Int(0)),
        cursor.set_index(s2),
        cursor.store(pt.Bytes("")),
        cursor.set_index(i3),
        cursor.store(pt.Int(0)),
        cursor.set_index(s3),
        cursor.store(pt.Bytes("")),
        pt.Int(42),
    )


@pt.Subroutine(pt.TealType.uint64)
def oldfac(n):
    return pt.If(n < pt.Int(2)).Then(pt.Int(1)).Else(n * oldfac(n - pt.Int(1)))


ISSUE_199_CASES = (
    sub_logcat_dynamic,
    swapper,
    wilt_the_stilt,
    sub_mixed,
    lots_o_vars,
    empty_scratches,
)


@pytest.mark.skipif(not STABLE_SLOT_GENERATION, reason="cf. #199")
@pytest.mark.parametrize("pt", ISSUE_199_CASES)
def test_teal_output_is_unchanged(pt):
    assert_new_v_old(pt, 6, "unchanged")


# #### pt.Subroutine Definitions for pass-by-ref guardrails testing #####
@pt.Subroutine(pt.TealType.uint64)
def ok(x):
    # not really ok at runtime... but should be ok at compile time
    return ok(x)


@pt.Subroutine(pt.TealType.uint64)
def ok_byref(x: pt.ScratchVar):
    return pt.Int(42)


@pt.Subroutine(pt.TealType.uint64)
def ok_indirect1(x):
    return ok_indirect2(x)


@pt.Subroutine(pt.TealType.uint64)
def ok_indirect2(x):
    return ok_indirect1(x)


@pt.Subroutine(pt.TealType.uint64)
def not_ok(x: pt.ScratchVar):
    # not ok both at compile and runtime
    return not_ok(x)


@pt.Subroutine(pt.TealType.uint64)
def not_ok_indirect1(x: pt.ScratchVar):
    return not_ok_indirect2(x)


@pt.Subroutine(pt.TealType.uint64)
def not_ok_indirect2(x: pt.ScratchVar):
    return not_ok_indirect1(x)


"""
Complex subroutine graph example:

a --> b --> a (loop)
        --> e
        --> f
    --> *c--> g --> a (loop)
        --> h
    --> d --> d (loop)
        --> i
        --> j

*c - this is the only "pass by-ref" subroutine
Expect the following error path: c-->g-->a-->c
"""


@pt.Subroutine(pt.TealType.uint64)
def a(x):
    tmp = pt.ScratchVar(pt.TealType.uint64)
    return pt.Seq(tmp.store(x), b(x) + c(tmp) + d(x))


@pt.Subroutine(pt.TealType.uint64)
def b(x):
    return a(x) + e(x) * f(x)


@pt.Subroutine(pt.TealType.uint64)
def c(x: pt.ScratchVar):
    return g(pt.Int(42)) - h(x.load())


@pt.Subroutine(pt.TealType.uint64)
def d(x):
    return d(x) + i(pt.Int(11)) * j(x)


@pt.Subroutine(pt.TealType.uint64)
def e(x):
    return pt.Int(42)


@pt.Subroutine(pt.TealType.uint64)
def f(x):
    return pt.Int(42)


@pt.Subroutine(pt.TealType.uint64)
def g(x):
    return a(pt.Int(17))


@pt.Subroutine(pt.TealType.uint64)
def h(x):
    return pt.Int(42)


@pt.Subroutine(pt.TealType.uint64)
def i(x):
    return pt.Int(42)


@pt.Subroutine(pt.TealType.uint64)
def j(x):
    return pt.Int(42)


@pt.Subroutine(pt.TealType.none)
def tally(n, result: pt.ScratchVar):
    return (
        pt.If(n == pt.Int(0))
        .Then(result.store(pt.Bytes("")))
        .Else(
            pt.Seq(
                tally(n - pt.Int(1), result),
                result.store(pt.Concat(result.load(), pt.Bytes("a"))),
            )
        )
    )


@pt.Subroutine(pt.TealType.none)
def factorial_BAD(n: pt.ScratchVar):
    tmp = pt.ScratchVar(pt.TealType.uint64)
    return (
        pt.If(n.load() <= pt.Int(1))
        .Then(n.store(pt.Int(1)))
        .Else(
            pt.Seq(
                tmp.store(n.load() - pt.Int(1)),
                factorial_BAD(tmp),
                n.store(n.load() * tmp.load()),
            )
        )
    )


@pt.Subroutine(pt.TealType.none)
def factorial(n: pt.ScratchVar):
    tmp = pt.ScratchVar(pt.TealType.uint64)
    return (
        pt.If(n.load() <= pt.Int(1))
        .Then(n.store(pt.Int(1)))
        .Else(
            pt.Seq(
                tmp.store(n.load()),
                n.store(n.load() - pt.Int(1)),
                factorial(n),
                n.store(n.load() * tmp.load()),
            )
        )
    )


@pt.Subroutine(pt.TealType.none)
def plus_one(n: pt.ScratchVar):
    tmp = pt.ScratchVar(pt.TealType.uint64)
    return (
        pt.If(n.load() == pt.Int(0))
        .Then(n.store(pt.Int(1)))
        .Else(
            pt.Seq(
                tmp.store(n.load() - pt.Int(1)),
                plus_one(tmp),
                n.store(tmp.load() + pt.Int(1)),
            )
        )
    )


def make_creatable_factory(approval):
    """
    Wrap a pyteal program with code that:
    * approves immediately in the case of app creation (appId == 0)
    * runs the original code otherwise
    """

    def func():
        return (
            pt.If(pt.Txn.application_id() == pt.Int(0)).Then(pt.Int(1)).Else(approval())
        )

    func.__name__ = approval.__name__
    return func


def fac_by_ref():
    n = pt.ScratchVar(pt.TealType.uint64)
    return pt.Seq(
        n.store(pt.Int(10)),
        factorial(n),
        n.load(),
    )


def fac_by_ref_BAD():
    n = pt.ScratchVar(pt.TealType.uint64)
    return pt.Seq(
        n.store(pt.Int(10)),
        factorial_BAD(n),
        n.load(),
    )


# Proved correct via blackbox testing, but BANNING for now
def fac_by_ref_args():
    n = pt.ScratchVar(pt.TealType.uint64)
    return pt.Seq(
        pt.If(
            pt.Or(
                pt.App.id() == pt.Int(0),
                pt.Txn.application_args.length() == pt.Int(0),
            )
        )
        .Then(pt.Int(1))
        .Else(
            pt.Seq(
                n.store(pt.Btoi(pt.Txn.application_args[0])),
                factorial(n),
                n.load(),
            )
        )
    )


def tallygo():
    result = pt.ScratchVar(pt.TealType.bytes)
    # pt.If-Then is a hook for creating + opting in without providing any args
    return (
        pt.If(
            pt.Or(
                pt.App.id() == pt.Int(0), pt.Txn.application_args.length() == pt.Int(0)
            )
        )
        .Then(pt.Int(1))
        .Else(
            pt.Seq(
                result.store(pt.Bytes("dummy")),
                tally(pt.Int(4), result),
                pt.Btoi(result.load()),
            )
        )
    )


TESTABLE_CASES = [(oldfac, [pt.TealType.uint64])]


# ---- Approval PyTEAL Expressions (COPACETIC) ---- #

approval_ok = ok(pt.Int(42))

x_scratchvar = pt.ScratchVar(pt.TealType.uint64)

approval_ok_byref = pt.Seq(x_scratchvar.store(pt.Int(42)), ok_byref(x_scratchvar))

approval_ok_indirect = ok_indirect1(pt.Int(42))

# ---- BANNED Approval PyTEAL Expressions (wrapped in a function) ---- #


@pt.Subroutine(pt.TealType.none)
def subr_string_mult(s: pt.ScratchVar, n):
    tmp = pt.ScratchVar(pt.TealType.bytes)
    return (
        pt.If(n == pt.Int(0))
        .Then(s.store(pt.Bytes("")))
        .Else(
            pt.Seq(
                tmp.store(s.load()),
                subr_string_mult(s, n - pt.Int(1)),
                s.store(pt.Concat(s.load(), tmp.load())),
            )
        )
    )


def string_mult():
    s = pt.ScratchVar(pt.TealType.bytes)
    return pt.Seq(
        s.store(pt.Txn.application_args[0]),
        subr_string_mult(s, pt.Btoi(pt.Txn.application_args[1])),
        pt.Log(s.load()),
        pt.Int(100),
    )


def approval_not_ok():
    return pt.Seq(x_scratchvar.store(pt.Int(42)), not_ok(x_scratchvar))


def approval_not_ok_indirect():
    return pt.Seq(x_scratchvar.store(pt.Int(42)), not_ok_indirect1(x_scratchvar))


def approval_its_complicated():
    return a(pt.Int(42))


def increment():
    n = pt.ScratchVar(pt.TealType.uint64)
    return pt.Seq(n.store(pt.Int(4)), plus_one(n), pt.Int(1))


COPACETIC_APPROVALS = [approval_ok, approval_ok_byref, approval_ok_indirect]


@pytest.mark.parametrize("approval", COPACETIC_APPROVALS)
def test_pass_by_ref_guardrails_COPACETIC(approval):
    assert pt.compileTeal(approval, pt.Mode.Application, version=6)


ILLEGAL_APPROVALS = {
    approval_not_ok: "not_ok()-->not_ok()",
    approval_not_ok_indirect: "not_ok_indirect1()-->not_ok_indirect2()-->not_ok_indirect1()",
    approval_its_complicated: "c()-->g()-->a()-->c()",
    fac_by_ref: "factorial()-->factorial()",
    fac_by_ref_args: "factorial()-->factorial()",
    fac_by_ref_BAD: "factorial_BAD()-->factorial_BAD()",
    increment: "plus_one()-->plus_one()",
    string_mult: "subr_string_mult()-->subr_string_mult()",
    tallygo: "tally()-->tally()",
}


@pytest.mark.parametrize("approval_func, suffix", ILLEGAL_APPROVALS.items())
def test_pass_by_ref_guardrails_BANNED(approval_func, suffix):
    with pytest.raises(pt.TealInputError) as err:
        pt.compileTeal(approval_func(), pt.Mode.Application, version=6)

    prefix = "ScratchVar arguments not allowed in recursive subroutines, but a recursive call-path was detected: "
    assert f"{prefix}{suffix}" in str(err)


def should_it_work() -> pt.Expr:
    xs = [
        pt.ScratchVar(pt.TealType.uint64),
        pt.ScratchVar(pt.TealType.uint64),
    ]

    def store_initial_values():
        return [s.store(pt.Int(i + 1)) for i, s in enumerate(xs)]

    d = pt.DynamicScratchVar(pt.TealType.uint64)

    @pt.Subroutine(pt.TealType.none)
    def retrieve_and_increment(s: pt.ScratchVar):
        return pt.Seq(d.set_index(s), d.store(d.load() + pt.Int(1)))

    def asserts():
        return [pt.Assert(x.load() == pt.Int(i + 2)) for i, x in enumerate(xs)]

    return pt.Seq(
        pt.Seq(store_initial_values()),
        pt.Seq([retrieve_and_increment(x) for x in xs]),
        pt.Seq(asserts()),
        pt.Int(1),
    )


def test_cannot_set_index_with_dynamic():
    with pytest.raises(pt.TealInputError) as tie:
        pt.compileTeal(should_it_work(), pt.Mode.Application, version=6)

    assert (
        "Only allowed to use ScratchVar objects for setting indices, but was given a"
        in str(tie)
    )



from typing import List, TYPE_CHECKING, overload

from pyteal.types import TealType, require_type
from pyteal.errors import TealInputError
from pyteal.ir import TealSimpleBlock
from pyteal.ast.expr import Expr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class Seq(Expr):
    """A control flow expression to represent a sequence of expressions."""

    @overload
    def __init__(self, *exprs: Expr) -> None:
        pass

    @overload
    def __init__(self, exprs: List[Expr]) -> None:
        pass

    def __init__(self, *exprs):
        """
        __init__(*exprs: Expr) -> None
        __init__(exprs: List[Expr]) -> None

        Create a new Seq expression.

        The new Seq expression will take on the return value of the final expression in the sequence.

        Args:
            exprs: The expressions to include in this sequence. All expressions that are not the
                final one in this list must not return any values.

        Example:
            .. code-block:: python

                Seq([
                    App.localPut(Bytes("key"), Bytes("value")),
                    Int(1)
                ])
        """
        super().__init__()

        # Handle case where a list of expressions is provided
        if len(exprs) == 1 and isinstance(exprs[0], list):
            exprs = exprs[0]

        for i, expr in enumerate(exprs):
            if not isinstance(expr, Expr):
                raise TealInputError("{} is not a pyteal expression.".format(expr))
            if i + 1 < len(exprs):
                require_type(expr, TealType.none)

        self.args = exprs

    def __teal__(self, options: "CompileOptions"):
        start = TealSimpleBlock([])
        end = start
        for arg in self.args:
            argStart, argEnd = arg.__teal__(options)
            end.setNextBlock(argStart)
            end = argEnd
        return start, end

    def __str__(self):
        ret_str = "(Seq"
        for a in self.args:
            ret_str += " " + a.__str__()
        ret_str += ")"
        return ret_str

    def type_of(self):
        if len(self.args) == 0:
            return TealType.none
        return self.args[-1].type_of()

    def has_return(self):
        # this expression declares it has a return op only if its final expression has a return op
        # TODO: technically if ANY expression, not just the final one, returns true for has_return,
        # this could return true as well. But in that case all expressions after the one that
        # returns true for has_return is dead code, so it could be optimized away
        if len(self.args) == 0:
            return False
        return self.args[-1].has_return()


Seq.__module__ = "pyteal"


@overload
def _use_seq_if_multiple(exprs: list[Expr]) -> Expr:
    ...


@overload
def _use_seq_if_multiple(*exprs: Expr) -> Expr:
    ...


def _use_seq_if_multiple(*exprs):
    """If multiple expressions are provided, wrap them in a Seq expression."""

    # Guard against no expressions
    if len(exprs) == 0:
        raise TealInputError("Expressions cannot be empty.")

    # Handle case where a list of expressions is provided
    if len(exprs) == 1 and isinstance(exprs[0], list):
        exprs = exprs[0]

    if len(exprs) > 1:
        return Seq(*exprs)
    return exprs[0]



from typing import NamedTuple
from enum import Enum, Flag, auto


class Mode(Flag):
    """Enum of program running modes."""

    Signature = auto()
    Application = auto()


Mode.__module__ = "pyteal"

OpType = NamedTuple("OpType", [("value", str), ("mode", Mode), ("min_version", int)])


class Op(Enum):
    """Enum of program opcodes."""

    def __str__(self) -> str:
        return self.value.value

    @property
    def mode(self) -> Mode:
        """Get the modes where this op is available."""
        return self.value.mode

    @property
    def min_version(self) -> int:
        """Get the minimum version where this op is available."""
        return self.value.min_version

    # fmt: off
    # meta
    comment             = OpType("//",                  Mode.Signature | Mode.Application, 0)
    # avm 
    err                 = OpType("err",                 Mode.Signature | Mode.Application, 2)
    sha256              = OpType("sha256",              Mode.Signature | Mode.Application, 2)
    keccak256           = OpType("keccak256",           Mode.Signature | Mode.Application, 2)
    sha512_256          = OpType("sha512_256",          Mode.Signature | Mode.Application, 2)
    ed25519verify       = OpType("ed25519verify",       Mode.Signature | Mode.Application, 2)
    add                 = OpType("+",                   Mode.Signature | Mode.Application, 2)
    minus               = OpType("-",                   Mode.Signature | Mode.Application, 2)
    div                 = OpType("/",                   Mode.Signature | Mode.Application, 2)
    mul                 = OpType("*",                   Mode.Signature | Mode.Application, 2)
    lt                  = OpType("<",                   Mode.Signature | Mode.Application, 2)
    gt                  = OpType(">",                   Mode.Signature | Mode.Application, 2)
    le                  = OpType("<=",                  Mode.Signature | Mode.Application, 2)
    ge                  = OpType(">=",                  Mode.Signature | Mode.Application, 2)
    logic_and           = OpType("&&",                  Mode.Signature | Mode.Application, 2)
    logic_or            = OpType("||",                  Mode.Signature | Mode.Application, 2)
    eq                  = OpType("==",                  Mode.Signature | Mode.Application, 2)
    neq                 = OpType("!=",                  Mode.Signature | Mode.Application, 2)
    logic_not           = OpType("!",                   Mode.Signature | Mode.Application, 2)
    len                 = OpType("len",                 Mode.Signature | Mode.Application, 2)
    itob                = OpType("itob",                Mode.Signature | Mode.Application, 2)
    btoi                = OpType("btoi",                Mode.Signature | Mode.Application, 2)
    mod                 = OpType("%",                   Mode.Signature | Mode.Application, 2)
    bitwise_or          = OpType("|",                   Mode.Signature | Mode.Application, 2)
    bitwise_and         = OpType("&",                   Mode.Signature | Mode.Application, 2)
    bitwise_xor         = OpType("^",                   Mode.Signature | Mode.Application, 2)
    bitwise_not         = OpType("~",                   Mode.Signature | Mode.Application, 2)
    mulw                = OpType("mulw",                Mode.Signature | Mode.Application, 2)
    addw                = OpType("addw",                Mode.Signature | Mode.Application, 2)
    intcblock           = OpType("intcblock",           Mode.Signature | Mode.Application, 2)
    intc                = OpType("intc",                Mode.Signature | Mode.Application, 2)
    intc_0              = OpType("intc_0",              Mode.Signature | Mode.Application, 2)
    intc_1              = OpType("intc_1",              Mode.Signature | Mode.Application, 2)
    intc_2              = OpType("intc_2",              Mode.Signature | Mode.Application, 2)
    intc_3              = OpType("intc_3",              Mode.Signature | Mode.Application, 2)
    int                 = OpType("int",                 Mode.Signature | Mode.Application, 2)
    bytecblock          = OpType("bytecblock",          Mode.Signature | Mode.Application, 2)
    bytec               = OpType("bytec",               Mode.Signature | Mode.Application, 2)
    bytec_0             = OpType("bytec_0",             Mode.Signature | Mode.Application, 2)
    bytec_1             = OpType("bytec_1",             Mode.Signature | Mode.Application, 2)
    bytec_2             = OpType("bytec_2",             Mode.Signature | Mode.Application, 2)
    bytec_3             = OpType("bytec_3",             Mode.Signature | Mode.Application, 2)
    byte                = OpType("byte",                Mode.Signature | Mode.Application, 2)
    addr                = OpType("addr",                Mode.Signature | Mode.Application, 2)
    method_signature    = OpType("method",              Mode.Signature | Mode.Application, 2)
    arg                 = OpType("arg",                 Mode.Signature,                    2)
    txn                 = OpType("txn",                 Mode.Signature | Mode.Application, 2)
    global_             = OpType("global",              Mode.Signature | Mode.Application, 2)
    gtxn                = OpType("gtxn",                Mode.Signature | Mode.Application, 2)
    load                = OpType("load",                Mode.Signature | Mode.Application, 2)
    store               = OpType("store",               Mode.Signature | Mode.Application, 2)
    txna                = OpType("txna",                Mode.Signature | Mode.Application, 2)
    gtxna               = OpType("gtxna",               Mode.Signature | Mode.Application, 2)
    bnz                 = OpType("bnz",                 Mode.Signature | Mode.Application, 2)
    bz                  = OpType("bz",                  Mode.Signature | Mode.Application, 2)
    b                   = OpType("b",                   Mode.Signature | Mode.Application, 2)
    return_             = OpType("return",              Mode.Signature | Mode.Application, 2)
    pop                 = OpType("pop",                 Mode.Signature | Mode.Application, 2)
    dup                 = OpType("dup",                 Mode.Signature | Mode.Application, 2)
    dup2                = OpType("dup2",                Mode.Signature | Mode.Application, 2)
    concat              = OpType("concat",              Mode.Signature | Mode.Application, 2)
    substring           = OpType("substring",           Mode.Signature | Mode.Application, 2)
    substring3          = OpType("substring3",          Mode.Signature | Mode.Application, 2)
    balance             = OpType("balance",             Mode.Application,                  2)
    app_opted_in        = OpType("app_opted_in",        Mode.Application,                  2)
    app_local_get       = OpType("app_local_get",       Mode.Application,                  2)
    app_local_get_ex    = OpType("app_local_get_ex",    Mode.Application,                  2)
    app_global_get      = OpType("app_global_get",      Mode.Application,                  2)
    app_global_get_ex   = OpType("app_global_get_ex",   Mode.Application,                  2)
    app_local_put       = OpType("app_local_put",       Mode.Application,                  2)
    app_global_put      = OpType("app_global_put",      Mode.Application,                  2)
    app_local_del       = OpType("app_local_del",       Mode.Application,                  2)
    app_global_del      = OpType("app_global_del",      Mode.Application,                  2)
    asset_holding_get   = OpType("asset_holding_get",   Mode.Application,                  2)
    asset_params_get    = OpType("asset_params_get",    Mode.Application,                  2)
    gtxns               = OpType("gtxns",               Mode.Signature | Mode.Application, 3)
    gtxnsa              = OpType("gtxnsa",              Mode.Signature | Mode.Application, 3)
    assert_             = OpType("assert",              Mode.Signature | Mode.Application, 3)
    dig                 = OpType("dig",                 Mode.Signature | Mode.Application, 3)
    swap                = OpType("swap",                Mode.Signature | Mode.Application, 3)
    select              = OpType("select",              Mode.Signature | Mode.Application, 3)
    getbit              = OpType("getbit",              Mode.Signature | Mode.Application, 3)
    setbit              = OpType("setbit",              Mode.Signature | Mode.Application, 3)
    getbyte             = OpType("getbyte",             Mode.Signature | Mode.Application, 3)
    setbyte             = OpType("setbyte",             Mode.Signature | Mode.Application, 3)
    min_balance         = OpType("min_balance",         Mode.Application,                  3)
    pushbytes           = OpType("pushbytes",           Mode.Signature | Mode.Application, 3)
    pushint             = OpType("pushint",             Mode.Signature | Mode.Application, 3)
    shl                 = OpType("shl",                 Mode.Signature | Mode.Application, 4)
    shr                 = OpType("shr",                 Mode.Signature | Mode.Application, 4)
    sqrt                = OpType("sqrt",                Mode.Signature | Mode.Application, 4)
    bitlen              = OpType("bitlen",              Mode.Signature | Mode.Application, 4)
    exp                 = OpType("exp",                 Mode.Signature | Mode.Application, 4)
    divmodw             = OpType("divmodw",             Mode.Signature | Mode.Application, 4)
    expw                = OpType("expw",                Mode.Signature | Mode.Application, 4)
    b_add               = OpType("b+",                  Mode.Signature | Mode.Application, 4)
    b_minus             = OpType("b-",                  Mode.Signature | Mode.Application, 4)
    b_div               = OpType("b/",                  Mode.Signature | Mode.Application, 4)
    b_mul               = OpType("b*",                  Mode.Signature | Mode.Application, 4)
    b_lt                = OpType("b<",                  Mode.Signature | Mode.Application, 4)
    b_gt                = OpType("b>",                  Mode.Signature | Mode.Application, 4)
    b_le                = OpType("b<=",                 Mode.Signature | Mode.Application, 4)
    b_ge                = OpType("b>=",                 Mode.Signature | Mode.Application, 4)
    b_eq                = OpType("b==",                 Mode.Signature | Mode.Application, 4)
    b_neq               = OpType("b!=",                 Mode.Signature | Mode.Application, 4)
    b_mod               = OpType("b%",                  Mode.Signature | Mode.Application, 4)
    b_or                = OpType("b|",                  Mode.Signature | Mode.Application, 4)
    b_and               = OpType("b&",                  Mode.Signature | Mode.Application, 4)
    b_xor               = OpType("b^",                  Mode.Signature | Mode.Application, 4)
    b_not               = OpType("b~",                  Mode.Signature | Mode.Application, 4)
    bzero               = OpType("bzero",               Mode.Signature | Mode.Application, 4)
    gload               = OpType("gload",               Mode.Application,                  4)
    gloads              = OpType("gloads",              Mode.Application,                  4)
    gaid                = OpType("gaid",                Mode.Application,                  4)
    gaids               = OpType("gaids",               Mode.Application,                  4)
    callsub             = OpType("callsub",             Mode.Signature | Mode.Application, 4)
    retsub              = OpType("retsub",              Mode.Signature | Mode.Application, 4)
    ecdsa_verify        = OpType("ecdsa_verify",        Mode.Signature | Mode.Application, 5)
    ecdsa_pk_decompress = OpType("ecdsa_pk_decompress", Mode.Signature | Mode.Application, 5)
    ecdsa_pk_recover    = OpType("ecdsa_pk_recover",    Mode.Signature | Mode.Application, 5)
    loads               = OpType("loads",               Mode.Signature | Mode.Application, 5)
    stores              = OpType("stores",              Mode.Signature | Mode.Application, 5)
    cover               = OpType("cover",               Mode.Signature | Mode.Application, 5)
    uncover             = OpType("uncover",             Mode.Signature | Mode.Application, 5)
    extract             = OpType("extract",             Mode.Signature | Mode.Application, 5)
    extract3            = OpType("extract3",            Mode.Signature | Mode.Application, 5)
    extract_uint16      = OpType("extract_uint16",      Mode.Signature | Mode.Application, 5)
    extract_uint32      = OpType("extract_uint32",      Mode.Signature | Mode.Application, 5)
    extract_uint64      = OpType("extract_uint64",      Mode.Signature | Mode.Application, 5)
    app_params_get      = OpType("app_params_get",      Mode.Application,                  5)
    log                 = OpType("log",                 Mode.Application,                  5)
    itxn_begin          = OpType("itxn_begin",          Mode.Application,                  5)
    itxn_field          = OpType("itxn_field",          Mode.Application,                  5)
    itxn_submit         = OpType("itxn_submit",         Mode.Application,                  5)
    itxn                = OpType("itxn",                Mode.Application,                  5)
    itxna               = OpType("itxna",               Mode.Application,                  5)
    txnas               = OpType("txnas",               Mode.Signature | Mode.Application, 5)
    gtxnas              = OpType("gtxnas",              Mode.Signature | Mode.Application, 5)
    gtxnsas             = OpType("gtxnsas",             Mode.Signature | Mode.Application, 5)
    args                = OpType("args",                Mode.Signature,                    5)
    bsqrt               = OpType("bsqrt",               Mode.Signature | Mode.Application, 6)
    divw                = OpType("divw",                Mode.Signature | Mode.Application, 6)
    itxn_next           = OpType("itxn_next",           Mode.Application,                  6)
    itxnas              = OpType("itxnas",              Mode.Application,                  6)
    gitxn               = OpType("gitxn",               Mode.Application,                  6)
    gitxna              = OpType("gitxna",              Mode.Application,                  6)
    gitxnas             = OpType("gitxnas",             Mode.Application,                  6)
    gloadss             = OpType("gloadss",             Mode.Application,                  6)
    acct_params_get     = OpType("acct_params_get",     Mode.Application,                  6)
    replace2            = OpType("replace2",            Mode.Signature | Mode.Application, 7)
    replace3            = OpType("replace3",            Mode.Signature | Mode.Application, 7)
    base64_decode       = OpType("base64_decode",       Mode.Signature | Mode.Application, 7)
    json_ref            = OpType("json_ref",            Mode.Signature | Mode.Application, 7)
    ed25519verify_bare  = OpType("ed25519verify_bare",  Mode.Signature | Mode.Application, 7)
    sha3_256            = OpType("sha3_256",            Mode.Signature | Mode.Application, 7)
    vrf_verify          = OpType("vrf_verify",          Mode.Signature | Mode.Application, 7)
    block               = OpType("block",               Mode.Signature | Mode.Application, 7)
    # fmt: on


Op.__module__ = "pyteal"



import pytest

import pyteal as pt

options = pt.CompileOptions()


def test_int():
    values = [0, 1, 8, 232323, 2**64 - 1]

    for value in values:
        expr = pt.Int(value)
        assert expr.type_of() == pt.TealType.uint64

        expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.int, value)])

        actual, _ = expr.__teal__(options)

        assert actual == expected


def test_int_invalid():
    with pytest.raises(pt.TealInputError):
        pt.Int(6.7)

    with pytest.raises(pt.TealInputError):
        pt.Int(-1)

    with pytest.raises(pt.TealInputError):
        pt.Int(2**64)

    with pytest.raises(pt.TealInputError):
        pt.Int("0")


def test_enum_int():
    expr = pt.EnumInt("OptIn")
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.int, "OptIn")])

    actual, _ = expr.__teal__(options)

    assert actual == expected



from typing import cast, Union, TYPE_CHECKING

from pyteal.types import TealType, require_type
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.errors import TealInputError, verifyProgramVersion
from pyteal.config import MAX_GROUP_SIZE
from pyteal.ast.expr import Expr
from pyteal.ast.leafexpr import LeafExpr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class GeneratedID(LeafExpr):
    """An expression to obtain the ID of an asset or application created by another transaction in the current group."""

    def __init__(self, txnIndex: Union[int, Expr]) -> None:
        """Create an expression to extract the created ID from a transaction in the current group.

        Requires program version 4 or higher. This operation is only permitted in application mode.

        Args:
            txnIndex: The index of the transaction from which the created ID should be obtained.
                This index may be a Python int, or it may be a PyTeal expression that evaluates at runtime.
                If it's an expression, it must evaluate to a uint64.
                In all cases, the index must be less than the index of the current transaction.
        """
        super().__init__()
        if type(txnIndex) is int:
            if txnIndex < 0 or txnIndex >= MAX_GROUP_SIZE:
                raise TealInputError(
                    "Invalid transaction index {}, shoud be in [0, {})".format(
                        txnIndex, MAX_GROUP_SIZE
                    )
                )
        else:
            require_type(cast(Expr, txnIndex), TealType.uint64)
        self.txnIndex = txnIndex

    def __str__(self):
        return "(Gaid {})".format(self.txnIndex)

    def __teal__(self, options: "CompileOptions"):
        verifyProgramVersion(
            Op.gaid.min_version,
            options.version,
            "Program version too low to use Gaid expression",
        )

        if type(self.txnIndex) is int:
            op = TealOp(self, Op.gaid, self.txnIndex)
            return TealBlock.FromOp(options, op)

        op = TealOp(self, Op.gaids)
        return TealBlock.FromOp(options, op, cast(Expr, self.txnIndex))

    def type_of(self):
        return TealType.uint64


GeneratedID.__module__ = "pyteal"



from typing import List, Tuple, Set, Dict, Optional, cast

from pyteal.compiler.optimizer import OptimizeOptions, apply_global_optimizations

from pyteal.types import TealType
from pyteal.ast import (
    Expr,
    Return,
    Seq,
    SubroutineDefinition,
    SubroutineDeclaration,
)
from pyteal.ir import Mode, Op, TealComponent, TealOp, TealBlock, TealSimpleBlock
from pyteal.errors import TealInputError, TealInternalError

from pyteal.compiler.sort import sortBlocks
from pyteal.compiler.flatten import flattenBlocks, flattenSubroutines
from pyteal.compiler.scratchslots import (
    assignScratchSlotsToSubroutines,
    collect_unoptimized_slots,
)
from pyteal.compiler.subroutines import (
    spillLocalSlotsDuringRecursion,
    resolveSubroutines,
)
from pyteal.compiler.constants import createConstantBlocks

MAX_PROGRAM_VERSION = 7
MIN_PROGRAM_VERSION = 2
DEFAULT_PROGRAM_VERSION = MIN_PROGRAM_VERSION


"""Deprecated. Use MAX_PROGRAM_VERSION instead."""
MAX_TEAL_VERSION = MAX_PROGRAM_VERSION
"""Deprecated. Use MIN_PROGRAM_VERSION instead."""
MIN_TEAL_VERSION = MIN_PROGRAM_VERSION
"""Deprecated. Use DEFAULT_PROGRAM_VERSION instead."""
DEFAULT_TEAL_VERSION = DEFAULT_PROGRAM_VERSION


class CompileOptions:
    def __init__(
        self,
        *,
        mode: Mode = Mode.Signature,
        version: int = DEFAULT_PROGRAM_VERSION,
        optimize: OptimizeOptions = None,
    ) -> None:
        self.mode = mode
        self.version = version
        self.optimize = optimize if optimize is not None else OptimizeOptions()

        self.currentSubroutine: Optional[SubroutineDefinition] = None

        self.breakBlocksStack: List[List[TealSimpleBlock]] = []
        self.continueBlocksStack: List[List[TealSimpleBlock]] = []

    def setSubroutine(self, subroutine: Optional[SubroutineDefinition]) -> None:
        self.currentSubroutine = subroutine

    def enterLoop(self) -> None:
        self.breakBlocksStack.append([])
        self.continueBlocksStack.append([])

    def isInLoop(self) -> bool:
        return len(self.breakBlocksStack) != 0

    def addLoopBreakBlock(self, block: TealSimpleBlock) -> None:
        if len(self.breakBlocksStack) == 0:
            raise TealInternalError("Cannot add break block when no loop is active")
        self.breakBlocksStack[-1].append(block)

    def addLoopContinueBlock(self, block: TealSimpleBlock) -> None:
        if len(self.continueBlocksStack) == 0:
            raise TealInternalError("Cannot add continue block when no loop is active")
        self.continueBlocksStack[-1].append(block)

    def exitLoop(self) -> Tuple[List[TealSimpleBlock], List[TealSimpleBlock]]:
        if len(self.breakBlocksStack) == 0 or len(self.continueBlocksStack) == 0:
            raise TealInternalError("Cannot exit loop when no loop is active")
        return self.breakBlocksStack.pop(), self.continueBlocksStack.pop()


def verifyOpsForVersion(teal: List[TealComponent], version: int):
    """Verify that all TEAL operations are allowed in the specified version.

    Args:
        teal: Code to check.
        mode: The version to check against.

    Raises:
        TealInputError: if teal contains an operation not allowed in version.
    """
    for stmt in teal:
        if isinstance(stmt, TealOp):
            op = stmt.getOp()
            if op.min_version > version:
                raise TealInputError(
                    "Op not supported in program version {}: {}. Minimum required version is {}".format(
                        version, op, op.min_version
                    )
                )


def verifyOpsForMode(teal: List[TealComponent], mode: Mode):
    """Verify that all TEAL operations are allowed in mode.

    Args:
        teal: Code to check.
        mode: The mode to check against.

    Raises:
        TealInputError: if teal contains an operation not allowed in mode.
    """
    for stmt in teal:
        if isinstance(stmt, TealOp):
            op = stmt.getOp()
            if not op.mode & mode:
                raise TealInputError(
                    "Op not supported in {} mode: {}".format(mode.name, op)
                )


def compileSubroutine(
    ast: Expr,
    options: CompileOptions,
    subroutineGraph: Dict[SubroutineDefinition, Set[SubroutineDefinition]],
    subroutine_start_blocks: Dict[Optional[SubroutineDefinition], TealBlock],
    subroutine_end_blocks: Dict[Optional[SubroutineDefinition], TealBlock],
) -> None:
    currentSubroutine = (
        cast(SubroutineDeclaration, ast).subroutine
        if isinstance(ast, SubroutineDeclaration)
        else None
    )

    if not ast.has_return():
        if ast.type_of() == TealType.none:
            ret_expr = Return()
            ret_expr.trace = ast.trace
            seq_expr = Seq([ast, ret_expr])
            seq_expr.trace = ret_expr.trace
            ast = seq_expr
        else:
            ret_expr = Return(ast)
            ret_expr.trace = ast.trace
            ast = ret_expr

    options.setSubroutine(currentSubroutine)

    start, end = ast.__teal__(options)
    start.addIncoming()
    start.validateTree()

    if (
        currentSubroutine is not None
        and currentSubroutine.get_declaration().deferred_expr is not None
    ):
        # this represents code that should be inserted before each retsub op
        deferred_expr = cast(Expr, currentSubroutine.get_declaration().deferred_expr)

        for block in TealBlock.Iterate(start):
            if not any(op.getOp() == Op.retsub for op in block.ops):
                continue

            if len(block.ops) != 1:
                # we expect all retsub ops to be in their own block at this point since
                # TealBlock.NormalizeBlocks has not yet been used
                raise TealInternalError(
                    f"Expected retsub to be the only op in the block, but there are {len(block.ops)} ops"
                )

            # we invoke __teal__ here and not outside of this loop because the same block cannot be
            # added in multiple places to the control flow graph
            deferred_start, deferred_end = deferred_expr.__teal__(options)
            deferred_start.addIncoming()
            deferred_start.validateTree()

            # insert deferred blocks between the previous block(s) and this one
            deferred_start.incoming = block.incoming
            block.incoming = [deferred_end]
            deferred_end.nextBlock = block

            for prev in deferred_start.incoming:
                prev.replaceOutgoing(block, deferred_start)

            if block is start:
                # this is the start block, replace start
                start = deferred_start

    start.validateTree()

    start = TealBlock.NormalizeBlocks(start)
    start.validateTree()

    subroutine_start_blocks[currentSubroutine] = start
    subroutine_end_blocks[currentSubroutine] = end

    referencedSubroutines: Set[SubroutineDefinition] = set()
    for block in TealBlock.Iterate(start):
        for stmt in block.ops:
            for subroutine in stmt.getSubroutines():
                referencedSubroutines.add(subroutine)

    if currentSubroutine is not None:
        subroutineGraph[currentSubroutine] = referencedSubroutines

    newSubroutines = referencedSubroutines - subroutine_start_blocks.keys()
    for subroutine in sorted(newSubroutines, key=lambda subroutine: subroutine.id):
        compileSubroutine(
            subroutine.get_declaration(),
            options,
            subroutineGraph,
            subroutine_start_blocks,
            subroutine_end_blocks,
        )


def sort_subroutine_blocks(
    subroutine_start_blocks: Dict[Optional[SubroutineDefinition], TealBlock],
    subroutine_end_blocks: Dict[Optional[SubroutineDefinition], TealBlock],
) -> Dict[Optional[SubroutineDefinition], List[TealComponent]]:
    subroutine_mapping: Dict[
        Optional[SubroutineDefinition], List[TealComponent]
    ] = dict()
    for subroutine, start in subroutine_start_blocks.items():
        order = sortBlocks(start, subroutine_end_blocks[subroutine])
        subroutine_mapping[subroutine] = flattenBlocks(order)

    return subroutine_mapping


def compileTeal(
    ast: Expr,
    mode: Mode,
    *,
    version: int = DEFAULT_PROGRAM_VERSION,
    assembleConstants: bool = False,
    optimize: OptimizeOptions = None,
) -> str:
    """Compile a PyTeal expression into TEAL assembly.

    Args:
        ast: The PyTeal expression to assemble.
        mode: The mode of the program to assemble. Must be Signature or Application.
        version (optional): The program version used to assemble the program. This will determine which
            expressions and fields are able to be used in the program and how expressions compile to
            TEAL opcodes. Defaults to 2 if not included.
        assembleConstants (optional): When true, the compiler will produce a program with fully
            assembled constants, rather than using the pseudo-ops `int`, `byte`, and `addr`. These
            constants will be assembled in the most space-efficient way, so enabling this may reduce
            the compiled program's size. Enabling this option requires a minimum program version of 3.
            Defaults to false.
        optimize (optional): OptimizeOptions that determine which optimizations will be applied.

    Returns:
        A TEAL assembly program compiled from the input expression.

    Raises:
        TealInputError: if an operation in ast is not supported by the supplied mode and version.
        TealInternalError: if an internal error is encounter during compilation.
    """
    if (
        not (MIN_PROGRAM_VERSION <= version <= MAX_PROGRAM_VERSION)
        or type(version) is not int
    ):
        raise TealInputError(
            "Unsupported program version: {}. Excepted an integer in the range [{}, {}]".format(
                version, MIN_PROGRAM_VERSION, MAX_PROGRAM_VERSION
            )
        )

    options = CompileOptions(mode=mode, version=version, optimize=optimize)

    subroutineGraph: Dict[SubroutineDefinition, Set[SubroutineDefinition]] = dict()
    subroutine_start_blocks: Dict[Optional[SubroutineDefinition], TealBlock] = dict()
    subroutine_end_blocks: Dict[Optional[SubroutineDefinition], TealBlock] = dict()
    compileSubroutine(
        ast, options, subroutineGraph, subroutine_start_blocks, subroutine_end_blocks
    )

    # note: optimizations are off by default, in which case, apply_global_optimizations
    # won't make any changes. Because the optimizer is invoked on a subroutine's
    # control flow graph, the optimizer requires context across block boundaries. This
    # is necessary for the dependency checking of local slots. Global slots, slots
    # used by DynamicScratchVar, and reserved slots are not optimized.
    if options.optimize.scratch_slots:
        options.optimize._skip_slots = collect_unoptimized_slots(
            subroutine_start_blocks
        )
        for start in subroutine_start_blocks.values():
            apply_global_optimizations(start, options.optimize)

    localSlotAssignments = assignScratchSlotsToSubroutines(subroutine_start_blocks)

    subroutineMapping: Dict[
        Optional[SubroutineDefinition], List[TealComponent]
    ] = sort_subroutine_blocks(subroutine_start_blocks, subroutine_end_blocks)

    spillLocalSlotsDuringRecursion(
        version, subroutineMapping, subroutineGraph, localSlotAssignments
    )

    subroutineLabels = resolveSubroutines(subroutineMapping)
    teal = flattenSubroutines(subroutineMapping, subroutineLabels)

    verifyOpsForVersion(teal, options.version)
    verifyOpsForMode(teal, options.mode)

    if assembleConstants:
        if version < 3:
            raise TealInternalError(
                "The minimum program version required to enable assembleConstants is 3. The current version is {}".format(
                    version
                )
            )
        teal = createConstantBlocks(teal)

    lines = ["#pragma version {}".format(version)]
    lines += [i.assemble() for i in teal]
    return "\n".join(lines)



from typing import Tuple, TYPE_CHECKING

from pyteal.types import TealType, require_type
from pyteal.errors import verifyProgramVersion
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.ast.expr import Expr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class TernaryExpr(Expr):
    """An expression with three arguments."""

    def __init__(
        self,
        op: Op,
        inputTypes: Tuple[TealType, TealType, TealType],
        outputType: TealType,
        firstArg: Expr,
        secondArg: Expr,
        thirdArg: Expr,
    ) -> None:
        super().__init__()
        require_type(firstArg, inputTypes[0])
        require_type(secondArg, inputTypes[1])
        require_type(thirdArg, inputTypes[2])

        self.op = op
        self.outputType = outputType
        self.firstArg = firstArg
        self.secondArg = secondArg
        self.thirdArg = thirdArg

    def __teal__(self, options: "CompileOptions"):
        verifyProgramVersion(
            self.op.min_version,
            options.version,
            "Program version too low to use op {}".format(self.op),
        )

        return TealBlock.FromOp(
            options, TealOp(self, self.op), self.firstArg, self.secondArg, self.thirdArg
        )

    def __str__(self):
        return "({} {} {} {})".format(
            self.op, self.firstArg, self.secondArg, self.thirdArg
        )

    def type_of(self):
        return self.outputType

    def has_return(self):
        return False


TernaryExpr.__module__ = "pyteal"


def Ed25519Verify(data: Expr, sig: Expr, key: Expr) -> TernaryExpr:
    """Verify the ed25519 signature of the concatenation ("ProgData" + hash_of_current_program + data).

    Args:
        data: The data signed by the public key. Must evaluate to bytes.
        sig: The proposed 64-byte signature of the concatenation ("ProgData" + hash_of_current_program + data).
            Must evaluate to bytes.
        key: The 32 byte public key that produced the signature. Must evaluate to bytes.
    """
    return TernaryExpr(
        Op.ed25519verify,
        (TealType.bytes, TealType.bytes, TealType.bytes),
        TealType.uint64,
        data,
        sig,
        key,
    )


def Ed25519Verify_Bare(data: Expr, sig: Expr, key: Expr) -> TernaryExpr:
    """Verify the ed25519 signature of the data against the public key.

    Args:
        data: The data signed by the public key. Must evaluate to bytes.
        sig: The proposed 64-byte signature of the data. Must evaluate to bytes.
        key: The 32 byte public key that produced the signature. Must evaluate to bytes.
    """
    return TernaryExpr(
        Op.ed25519verify_bare,
        (TealType.bytes, TealType.bytes, TealType.bytes),
        TealType.uint64,
        data,
        sig,
        key,
    )


def SetBit(value: Expr, index: Expr, newBitValue: Expr) -> TernaryExpr:
    """Set the bit value of an expression at a specific index.

    The meaning of index differs if value is an integer or a byte string.

    * For integers, bit indexing begins with low-order bits. For example, :code:`SetBit(Int(0), Int(4), Int(1))`
      yields the integer 16 (2^4). Any integer less than 64 is a valid index.

    * For byte strings, bit indexing begins at the first byte. For example, :code:`SetBit(Bytes("base16", "0x00"), Int(7), Int(1))`
      yields the byte string 0x01. Any integer less than 8*Len(value) is a valid index.

    Requires program version 3 or higher.

    Args:
        value: The value containing bits. Can evaluate to any type.
        index: The index of the bit to set. Must evaluate to uint64.
        newBitValue: The new bit value to set. Must evaluate to the integer 0 or 1.
    """
    return TernaryExpr(
        Op.setbit,
        (TealType.anytype, TealType.uint64, TealType.uint64),
        value.type_of(),
        value,
        index,
        newBitValue,
    )


def SetByte(value: Expr, index: Expr, newByteValue: Expr) -> TernaryExpr:
    """Set a single byte in a byte string from an integer value.

    Similar to SetBit, indexing begins at the first byte. For example, :code:`SetByte(Bytes("base16", "0x000000"), Int(0), Int(255))`
    yields the byte string 0xff0000.

    Requires program version 3 or higher.

    Args:
        value: The value containing the bytes. Must evaluate to bytes.
        index: The index of the byte to set. Must evaluate to an integer less than Len(value).
        newByteValue: The new byte value to set. Must evaluate to an integer less than 256.
    """
    return TernaryExpr(
        Op.setbyte,
        (TealType.bytes, TealType.uint64, TealType.uint64),
        TealType.bytes,
        value,
        index,
        newByteValue,
    )


def Divw(hi: Expr, lo: Expr, y: Expr) -> TernaryExpr:
    """
    Performs wide division by interpreting `hi` and `lo` as a uint128 value.

    Requires program version 6 or higher.

    Args:
        hi: Quotient's high 64 bits.  Must evaluate to uint64.
        lo: Quotient's low 64 bits.  Must evaluate to uint64.
        y: Divisor.  Must evaluate to uint64.

    """
    return TernaryExpr(
        Op.divw,
        (TealType.uint64, TealType.uint64, TealType.uint64),
        TealType.uint64,
        hi,
        lo,
        y,
    )



from typing import NamedTuple, List

import pyteal as pt

options = pt.CompileOptions()


def test_from_op_no_args():
    op = pt.TealOp(None, pt.Op.int, 1)

    expected = pt.TealSimpleBlock([op])

    actual, _ = pt.TealBlock.FromOp(options, op)

    assert actual == expected


def test_from_op_1_arg():
    op = pt.TealOp(None, pt.Op.pop)
    arg_1 = pt.Bytes("message")

    expected = pt.TealSimpleBlock([pt.TealOp(arg_1, pt.Op.byte, '"message"'), op])

    actual, _ = pt.TealBlock.FromOp(options, op, arg_1)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)
    actual.validateTree()

    assert actual == expected


def test_from_op_2_args():
    op = pt.TealOp(None, pt.Op.app_global_put)
    arg_1 = pt.Bytes("key")
    arg_2 = pt.Int(5)

    expected = pt.TealSimpleBlock(
        [pt.TealOp(arg_1, pt.Op.byte, '"key"'), pt.TealOp(arg_2, pt.Op.int, 5), op]
    )

    actual, _ = pt.TealBlock.FromOp(options, op, arg_1, arg_2)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)
    actual.validateTree()

    assert actual == expected


def test_from_op_3_args():
    op = pt.TealOp(None, pt.Op.app_local_put)
    arg_1 = pt.Int(0)
    arg_2 = pt.Bytes("key")
    arg_3 = pt.Int(1)
    arg_4 = pt.Int(2)
    arg_3_plus_4 = arg_3 + arg_4

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg_1, pt.Op.int, 0),
            pt.TealOp(arg_2, pt.Op.byte, '"key"'),
            pt.TealOp(arg_3, pt.Op.int, 1),
            pt.TealOp(arg_4, pt.Op.int, 2),
            pt.TealOp(arg_3_plus_4, pt.Op.add),
            op,
        ]
    )

    actual, _ = pt.TealBlock.FromOp(options, op, arg_1, arg_2, arg_3_plus_4)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)
    actual.validateTree()

    assert actual == expected


def test_iterate_single():
    block = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 1)])

    blocks = list(pt.TealBlock.Iterate(block))

    assert blocks == [block]


def test_iterate_sequence():
    block5 = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 5)])
    block4 = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 4)])
    block4.setNextBlock(block5)
    block3 = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 3)])
    block3.setNextBlock(block4)
    block2 = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 2)])
    block2.setNextBlock(block3)
    block1 = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 1)])
    block1.setNextBlock(block2)

    blocks = list(pt.TealBlock.Iterate(block1))

    assert blocks == [block1, block2, block3, block4, block5]


def test_iterate_branch():
    blockTrue = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"true"')])
    blockFalse = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"false"')])
    block = pt.TealConditionalBlock([pt.TealOp(None, pt.Op.int, 1)])
    block.setTrueBlock(blockTrue)
    block.setFalseBlock(blockFalse)

    blocks = list(pt.TealBlock.Iterate(block))

    assert blocks == [block, blockTrue, blockFalse]


def test_iterate_multiple_branch():
    blockTrueTrue = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"true true"')])
    blockTrueFalse = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"true false"')])
    blockTrueBranch = pt.TealConditionalBlock([])
    blockTrueBranch.setTrueBlock(blockTrueTrue)
    blockTrueBranch.setFalseBlock(blockTrueFalse)
    blockTrue = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"true"')])
    blockTrue.setNextBlock(blockTrueBranch)
    blockFalse = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"false"')])
    block = pt.TealConditionalBlock([pt.TealOp(None, pt.Op.int, 1)])
    block.setTrueBlock(blockTrue)
    block.setFalseBlock(blockFalse)

    blocks = list(pt.TealBlock.Iterate(block))

    assert blocks == [
        block,
        blockTrue,
        blockFalse,
        blockTrueBranch,
        blockTrueTrue,
        blockTrueFalse,
    ]


def test_iterate_branch_converge():
    blockEnd = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.return_)])
    blockTrue = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"true"')])
    blockTrue.setNextBlock(blockEnd)
    blockFalse = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"false"')])
    blockFalse.setNextBlock(blockEnd)
    block = pt.TealConditionalBlock([pt.TealOp(None, pt.Op.int, 1)])
    block.setTrueBlock(blockTrue)
    block.setFalseBlock(blockFalse)

    blocks = list(pt.TealBlock.Iterate(block))

    assert blocks == [block, blockTrue, blockFalse, blockEnd]


def test_normalize_single():
    original = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 1)])

    expected = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 1)])

    original.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(original)
    actual.validateTree()

    assert actual == expected


def test_normalize_sequence():
    block6 = pt.TealSimpleBlock([])
    block5 = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 5)])
    block5.setNextBlock(block6)
    block4 = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 4)])
    block4.setNextBlock(block5)
    block3 = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 3)])
    block3.setNextBlock(block4)
    block2 = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 2)])
    block2.setNextBlock(block3)
    block1 = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 1)])
    block1.setNextBlock(block2)

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.int, 2),
            pt.TealOp(None, pt.Op.int, 3),
            pt.TealOp(None, pt.Op.int, 4),
            pt.TealOp(None, pt.Op.int, 5),
        ]
    )

    block1.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(block1)
    actual.validateTree()

    assert actual == expected


def test_normalize_branch():
    blockTrueNext = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 4)])
    blockTrue = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"true"')])
    blockTrue.setNextBlock(blockTrueNext)
    blockFalse = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"false"')])
    blockBranch = pt.TealConditionalBlock([pt.TealOp(None, pt.Op.int, 1)])
    blockBranch.setTrueBlock(blockTrue)
    blockBranch.setFalseBlock(blockFalse)
    original = pt.TealSimpleBlock([])
    original.setNextBlock(blockBranch)

    expectedTrue = pt.TealSimpleBlock(
        [pt.TealOp(None, pt.Op.byte, '"true"'), pt.TealOp(None, pt.Op.int, 4)]
    )
    expectedFalse = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"false"')])
    expected = pt.TealConditionalBlock([pt.TealOp(None, pt.Op.int, 1)])
    expected.setTrueBlock(expectedTrue)
    expected.setFalseBlock(expectedFalse)

    original.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(original)
    actual.validateTree()

    assert actual == expected


def test_normalize_branch_converge():
    blockEnd = pt.TealSimpleBlock([])
    blockTrueNext = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 4)])
    blockTrueNext.setNextBlock(blockEnd)
    blockTrue = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"true"')])
    blockTrue.setNextBlock(blockTrueNext)
    blockFalse = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"false"')])
    blockFalse.setNextBlock(blockEnd)
    blockBranch = pt.TealConditionalBlock([pt.TealOp(None, pt.Op.int, 1)])
    blockBranch.setTrueBlock(blockTrue)
    blockBranch.setFalseBlock(blockFalse)
    original = pt.TealSimpleBlock([])
    original.setNextBlock(blockBranch)

    expectedEnd = pt.TealSimpleBlock([])
    expectedTrue = pt.TealSimpleBlock(
        [pt.TealOp(None, pt.Op.byte, '"true"'), pt.TealOp(None, pt.Op.int, 4)]
    )
    expectedTrue.setNextBlock(expectedEnd)
    expectedFalse = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"false"')])
    expectedFalse.setNextBlock(expectedEnd)
    expected = pt.TealConditionalBlock([pt.TealOp(None, pt.Op.int, 1)])
    expected.setTrueBlock(expectedTrue)
    expected.setFalseBlock(expectedFalse)

    original.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(original)
    actual.validateTree()

    assert actual == expected


def test_GetReferencedScratchSlots():
    a = pt.ScratchSlot()
    b = pt.ScratchSlot()
    c = pt.ScratchSlot()
    d = pt.ScratchSlot()

    end = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.load, d)])
    trueBranch = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.load, b)])
    trueBranch.setNextBlock(end)
    falseBranch = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.load, c)])
    falseBranch.setNextBlock(end)
    splitBranch = pt.TealConditionalBlock([pt.TealOp(None, pt.Op.load, a)])
    splitBranch.setTrueBlock(trueBranch)
    splitBranch.setFalseBlock(falseBranch)

    slotReferences = pt.TealBlock.GetReferencedScratchSlots(splitBranch)
    assert slotReferences == [a, b, c, d]


def test_MatchScratchSlotReferences():
    class MatchSlotReferenceTest(NamedTuple):
        actual: List[pt.ScratchSlot]
        expected: List[pt.ScratchSlot]
        match: bool

    a = pt.ScratchSlot()
    b = pt.ScratchSlot()
    c = pt.ScratchSlot()
    d = pt.ScratchSlot()

    tests: List[MatchSlotReferenceTest] = [
        MatchSlotReferenceTest(
            actual=[],
            expected=[],
            match=True,
        ),
        MatchSlotReferenceTest(
            actual=[a],
            expected=[],
            match=False,
        ),
        MatchSlotReferenceTest(
            actual=[a],
            expected=[a],
            match=True,
        ),
        MatchSlotReferenceTest(
            actual=[a],
            expected=[b],
            match=True,
        ),
        MatchSlotReferenceTest(
            actual=[a, a],
            expected=[a, a],
            match=True,
        ),
        MatchSlotReferenceTest(
            actual=[a, a],
            expected=[b, b],
            match=True,
        ),
        MatchSlotReferenceTest(
            actual=[a, b],
            expected=[a, b],
            match=True,
        ),
        MatchSlotReferenceTest(
            actual=[a, b],
            expected=[b, c],
            match=False,
        ),
        MatchSlotReferenceTest(
            actual=[a, b],
            expected=[c, d],
            match=True,
        ),
        MatchSlotReferenceTest(
            actual=[a, b, b, a, b],
            expected=[c, d, d, c, d],
            match=True,
        ),
        MatchSlotReferenceTest(
            actual=[a, b, b, a, b],
            expected=[a, d, d, a, d],
            match=True,
        ),
        MatchSlotReferenceTest(
            actual=[a, b, b, a, b],
            expected=[c, a, a, c, a],
            match=False,
        ),
        MatchSlotReferenceTest(
            actual=[a, b, b, a, b],
            expected=[c, d, d, c, c],
            match=False,
        ),
    ]

    for i, test in enumerate(tests):
        assert (
            pt.TealBlock.MatchScratchSlotReferences(test.actual, test.expected)
            == test.match
        ), "Test at index {} failed".format(i)



from typing import List

import pytest
import pyteal as pt

from algosdk.abi import ABIType
from pyteal import abi
from pyteal.ast.abi.util import substring_for_decoding
from pyteal.ast.abi.tuple import _encode_tuple
from pyteal.ast.abi.array_base_test import STATIC_TYPES, DYNAMIC_TYPES
from pyteal.ast.abi.type_test import ContainerType

options = pt.CompileOptions(version=5)


def test_DynamicArrayTypeSpec_init():
    for elementType in STATIC_TYPES:
        dynamicArrayType = abi.DynamicArrayTypeSpec(elementType)
        assert dynamicArrayType.value_type_spec() is elementType
        assert dynamicArrayType.is_length_dynamic()
        assert dynamicArrayType._stride() == elementType.byte_length_static()

    dynamicBytesType = abi.DynamicBytesTypeSpec()
    assert isinstance(dynamicBytesType.value_type_spec(), abi.ByteTypeSpec)
    assert dynamicBytesType.is_length_dynamic()
    assert dynamicBytesType._stride() == 1

    for elementType in DYNAMIC_TYPES:
        dynamicArrayType = abi.DynamicArrayTypeSpec(elementType)
        assert dynamicArrayType.value_type_spec() is elementType
        assert dynamicArrayType.is_length_dynamic()
        assert dynamicArrayType._stride() == 2


def test_DynamicArrayTypeSpec_str():
    for elementType in STATIC_TYPES + DYNAMIC_TYPES:
        dynamicArrayType = abi.DynamicArrayTypeSpec(elementType)
        assert str(dynamicArrayType) == "{}[]".format(elementType)

    assert str(abi.DynamicBytesTypeSpec()) == "byte[]"


def test_DynamicArrayTypeSpec_new_instance():
    for elementType in STATIC_TYPES + DYNAMIC_TYPES:
        dynamicArrayType = abi.DynamicArrayTypeSpec(elementType)
        instance = dynamicArrayType.new_instance()
        assert isinstance(instance, abi.DynamicArray)
        assert instance.type_spec() == dynamicArrayType

    dynamicBytesType = abi.DynamicBytesTypeSpec()
    instance = dynamicBytesType.new_instance()
    assert isinstance(instance, abi.DynamicBytes)
    assert instance.type_spec() == dynamicBytesType


def test_DynamicArrayTypeSpec_eq():
    for elementType in STATIC_TYPES + DYNAMIC_TYPES:
        dynamicArrayType = abi.DynamicArrayTypeSpec(elementType)
        assert dynamicArrayType == dynamicArrayType
        assert dynamicArrayType != abi.TupleTypeSpec(dynamicArrayType)

    dynamicBytesType = abi.DynamicBytesTypeSpec()
    assert dynamicBytesType == dynamicBytesType
    assert dynamicBytesType != abi.TupleTypeSpec(dynamicBytesType)


def test_DynamicArrayTypeSpec_is_dynamic():
    for elementType in STATIC_TYPES + DYNAMIC_TYPES:
        dynamicArrayType = abi.DynamicArrayTypeSpec(elementType)
        assert dynamicArrayType.is_dynamic()

    assert abi.DynamicBytesTypeSpec().is_dynamic()


def test_DynamicArrayTypeSpec_byte_length_static():
    for elementType in STATIC_TYPES + DYNAMIC_TYPES:
        dynamicArrayType = abi.DynamicArrayTypeSpec(elementType)
        with pytest.raises(ValueError):
            dynamicArrayType.byte_length_static()

    with pytest.raises(ValueError):
        abi.DynamicBytesTypeSpec().byte_length_static()


def test_DynamicArray_decode():
    encoded = pt.Bytes("encoded")
    dynamicArrayType = abi.DynamicArrayTypeSpec(abi.Uint64TypeSpec())
    for start_index in (None, pt.Int(1)):
        for end_index in (None, pt.Int(2)):
            for length in (None, pt.Int(3)):
                value = dynamicArrayType.new_instance()

                if end_index is not None and length is not None:
                    with pytest.raises(pt.TealInputError):
                        value.decode(
                            encoded,
                            start_index=start_index,
                            end_index=end_index,
                            length=length,
                        )
                    continue

                expr = value.decode(
                    encoded, start_index=start_index, end_index=end_index, length=length
                )
                assert expr.type_of() == pt.TealType.none
                assert not expr.has_return()

                expectedExpr = value.stored_value.store(
                    substring_for_decoding(
                        encoded,
                        start_index=start_index,
                        end_index=end_index,
                        length=length,
                    )
                )
                expected, _ = expectedExpr.__teal__(options)
                expected.addIncoming()
                expected = pt.TealBlock.NormalizeBlocks(expected)

                actual, _ = expr.__teal__(options)
                actual.addIncoming()
                actual = pt.TealBlock.NormalizeBlocks(actual)

                with pt.TealComponent.Context.ignoreExprEquality():
                    assert actual == expected


def test_DynamicArray_set_values():
    valuesToSet: List[abi.Uint64] = [
        [],
        [abi.Uint64()],
        [abi.Uint64() for _ in range(10)],
    ]

    dynamicArrayType = abi.DynamicArrayTypeSpec(abi.Uint64TypeSpec())
    for values in valuesToSet:
        value = dynamicArrayType.new_instance()
        expr = value.set(values)
        assert expr.type_of() == pt.TealType.none
        assert not expr.has_return()

        length_tmp = abi.Uint16()
        expectedExpr = value.stored_value.store(
            pt.Concat(
                pt.Seq(length_tmp.set(len(values)), length_tmp.encode()),
                _encode_tuple(values),
            )
        )
        expected, _ = expectedExpr.__teal__(options)
        expected.addIncoming()
        expected = pt.TealBlock.NormalizeBlocks(expected)

        actual, _ = expr.__teal__(options)
        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        with pt.TealComponent.Context.ignoreExprEquality():
            with pt.TealComponent.Context.ignoreScratchSlotEquality():
                assert actual == expected

        assert pt.TealBlock.MatchScratchSlotReferences(
            pt.TealBlock.GetReferencedScratchSlots(actual),
            pt.TealBlock.GetReferencedScratchSlots(expected),
        )


def test_DynamicArray_set_copy():
    dynamicArrayType = abi.DynamicArrayTypeSpec(abi.Uint64TypeSpec())
    value = dynamicArrayType.new_instance()
    otherArray = dynamicArrayType.new_instance()

    with pytest.raises(pt.TealInputError):
        value.set(abi.DynamicArray(abi.DynamicArrayTypeSpec(abi.Uint8TypeSpec())))

    with pytest.raises(pt.TealInputError):
        value.set(abi.Uint64())

    expr = value.set(otherArray)
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.load, otherArray.stored_value.slot),
            pt.TealOp(None, pt.Op.store, value.stored_value.slot),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_DynamicArray_set_computed():
    value = abi.DynamicArray(abi.DynamicArrayTypeSpec(abi.ByteTypeSpec()))
    computed = ContainerType(
        value.type_spec(), pt.Bytes("this should be a dynamic array")
    )
    expr = value.set(computed)
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.byte, '"this should be a dynamic array"'),
            pt.TealOp(None, pt.Op.store, value.stored_value.slot),
        ]
    )
    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = actual.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    with pytest.raises(pt.TealInputError):
        value.set(
            ContainerType(
                abi.DynamicArrayTypeSpec(abi.Uint16TypeSpec()),
                pt.Bytes("well i am trolling again"),
            )
        )


# AACS key recovery
BYTE_HEX_TEST_CASE = "09f911029d74e35bd84156c5635688c0"


BYTES_SET_TESTCASES = [
    bytes.fromhex(BYTE_HEX_TEST_CASE),
    bytearray.fromhex(BYTE_HEX_TEST_CASE),
]


@pytest.mark.parametrize("test_case", BYTES_SET_TESTCASES)
def test_DynamicBytes_set_py_bytes(test_case: bytes | bytearray):
    value = abi.DynamicBytes()

    expr = value.set(test_case)
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = actual.NormalizeBlocks(actual)

    length_encoding = ABIType.from_string("uint16").encode(len(test_case)).hex()

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.byte, "0x" + length_encoding + BYTE_HEX_TEST_CASE),
            pt.TealOp(None, pt.Op.store, value.stored_value.slot),
        ]
    )

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


@pytest.mark.parametrize("test_case", BYTES_SET_TESTCASES)
def test_DynamicBytes_set_expr(test_case: bytes | bytearray):
    value = abi.DynamicBytes()

    set_expr = pt.Concat(pt.Bytes(test_case), pt.Bytes(test_case))

    expr = value.set(set_expr)
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = actual.NormalizeBlocks(actual)

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.byte, "0x" + BYTE_HEX_TEST_CASE),
            pt.TealOp(None, pt.Op.byte, "0x" + BYTE_HEX_TEST_CASE),
            pt.TealOp(None, pt.Op.concat),
            pt.TealOp(None, pt.Op.store, value.stored_value.slot),
            pt.TealOp(None, pt.Op.load, value.stored_value.slot),
            pt.TealOp(None, pt.Op.len),
            pt.TealOp(None, pt.Op.itob),
            pt.TealOp(None, pt.Op.extract, 6, 0),
            pt.TealOp(None, pt.Op.load, value.stored_value.slot),
            pt.TealOp(None, pt.Op.concat),
            pt.TealOp(None, pt.Op.store, value.stored_value.slot),
        ]
    )

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_DynamicBytes_get():
    value = abi.DynamicBytes()

    expr = value.get()
    assert expr.type_of() == pt.TealType.bytes
    assert not expr.has_return()

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = actual.NormalizeBlocks(actual)

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.load, value.stored_value.slot),
            pt.TealOp(None, pt.Op.extract, 2, 0),
        ]
    )

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_DynamicArray_encode():
    dynamicArrayType = abi.DynamicArrayTypeSpec(abi.Uint64TypeSpec())
    value = dynamicArrayType.new_instance()
    expr = value.encode()
    assert expr.type_of() == pt.TealType.bytes
    assert not expr.has_return()

    expected = pt.TealSimpleBlock(
        [pt.TealOp(None, pt.Op.load, value.stored_value.slot)]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_DynamicArray_length():
    dynamicArrayType = abi.DynamicArrayTypeSpec(abi.Uint64TypeSpec())
    value = dynamicArrayType.new_instance()
    expr = value.length()
    assert expr.type_of() == pt.TealType.uint64
    assert not expr.has_return()

    length_tmp = abi.Uint16()
    expectedExpr = pt.Seq(length_tmp.decode(value.encode()), length_tmp.get())
    expected, _ = expectedExpr.__teal__(options)
    expected.addIncoming()
    expected = pt.TealBlock.NormalizeBlocks(expected)

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        with pt.TealComponent.Context.ignoreScratchSlotEquality():
            assert actual == expected

    assert pt.TealBlock.MatchScratchSlotReferences(
        pt.TealBlock.GetReferencedScratchSlots(actual),
        pt.TealBlock.GetReferencedScratchSlots(expected),
    )


def test_DynamicArray_getitem():
    dynamicArrayType = abi.DynamicArrayTypeSpec(abi.Uint64TypeSpec())
    value = dynamicArrayType.new_instance()

    for index in (0, 1, 2, 3, 1000):
        # dynamic indexes
        indexExpr = pt.Int(index)
        element = value[indexExpr]
        assert type(element) is abi.ArrayElement
        assert element.array is value
        assert element.index is indexExpr

    for index in (0, 1, 2, 3, 1000):
        # static indexes
        element = value[index]
        assert type(element) is abi.ArrayElement
        assert element.array is value
        assert type(element.index) is pt.Int
        assert element.index.value == index

    with pytest.raises(pt.TealInputError):
        value[-1]



import pytest

import pyteal as pt

avm6Options = pt.CompileOptions(version=6)
avm7Options = pt.CompileOptions(version=7)


def test_json_string():
    args = [pt.Bytes('{"foo":"bar"}'), pt.Bytes("foo")]
    expr = pt.JsonRef.as_string(*args)
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"{\\"foo\\":\\"bar\\"}"'),
            pt.TealOp(args[1], pt.Op.byte, '"foo"'),
            pt.TealOp(expr, pt.Op.json_ref, "JSONString"),
        ]
    )

    actual, _ = expr.__teal__(avm7Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm6Options)


def test_json_uint64():
    args = [pt.Bytes('{"foo":123456789}'), pt.Bytes("foo")]
    expr = pt.JsonRef.as_uint64(*args)
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"{\\"foo\\":123456789}"'),
            pt.TealOp(args[1], pt.Op.byte, '"foo"'),
            pt.TealOp(expr, pt.Op.json_ref, "JSONUint64"),
        ]
    )

    actual, _ = expr.__teal__(avm7Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm6Options)


def test_json_object():
    args = [pt.Bytes('{"foo":{"key": "value"}}'), pt.Bytes("foo")]
    expr = pt.JsonRef.as_object(*args)
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"{\\"foo\\":{\\"key\\": \\"value\\"}}"'),
            pt.TealOp(args[1], pt.Op.byte, '"foo"'),
            pt.TealOp(expr, pt.Op.json_ref, "JSONObject"),
        ]
    )

    actual, _ = expr.__teal__(avm7Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm6Options)


def test_json_ref_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.JsonRef.as_object(pt.Int(0), pt.Bytes("a"))

    with pytest.raises(pt.TealTypeError):
        pt.JsonRef.as_string(pt.Bytes("a"), pt.Int(0))



from pyteal.errors import TealInternalError


def escapeStr(s: str) -> str:
    """Escape a UTF-8 string for use in TEAL assembly.

    Args:
        s: A UTF-8 string to escape.

    Returns:
        An escaped version of the input string. This version will be surrounded in double quotes,
        all special characters (such as \\n) will be escaped with additional backslashes, and all
        Unicode characters beyond the latin-1 encoding will be encoded in hex escapes (e.g. \\xf0).
    """
    # The point of this conversion is to escape all special characters and turn all Unicode
    # characters into hex-escaped characters in the input string.
    #
    # The first step breaks up large Unicode characters into multiple UTF-8 hex characters:
    #     s_1 = s.encode("utf-8").decode("latin-1"), e.g. "\n 😀" => "\n ð\x9f\x98\x80"
    #
    # The next step escapes all special characters:
    #     s_1.encode("unicode-escape").decode("latin-1"), e.g. "\n ð\x9f\x98\x80" => "\\n \\xf0\\x9f\\x98\\x80"
    #
    # If we skipped the first step we would end up with Unicode codepoints instead of hex escaped
    # characters, which TEAL assembly cannot process:
    #     s.encode("unicode-escape").decode("latin-1"), e.g. "\n 😀" => "\\n \\U0001f600'"
    s = s.encode("utf-8").decode("latin-1").encode("unicode-escape").decode("latin-1")

    # Escape double quote characters (not covered by unicode-escape) but leave in single quotes
    s = s.replace('"', '\\"')

    # Surround string in double quotes
    return '"' + s + '"'


def unescapeStr(s: str) -> str:
    if len(s) < 2 or s[0] != '"' or s[-1] != '"':
        raise ValueError("Escaped string if of the wrong format")
    s = s[1:-1]
    s = s.replace('\\"', '"')
    s = s.encode("latin-1").decode("unicode-escape").encode("latin-1").decode("utf-8")
    return s


def correctBase32Padding(s: str) -> str:
    content = s.split("=")[0]
    trailing = len(content) % 8

    if trailing == 2:
        content += "=" * 6
    elif trailing == 4:
        content += "=" * 4
    elif trailing == 5:
        content += "=" * 3
    elif trailing == 7:
        content += "="
    elif trailing != 0:
        raise TealInternalError("Invalid base32 content")

    return content



from typing import Dict, Callable

import pytest

import pyteal as pt

fieldToMethod: Dict[pt.TxnField, Callable[[pt.TxnObject], pt.Expr]] = {
    pt.TxnField.sender: lambda txn: txn.sender(),
    pt.TxnField.fee: lambda txn: txn.fee(),
    pt.TxnField.first_valid: lambda txn: txn.first_valid(),
    pt.TxnField.first_valid_time: lambda txn: txn.first_valid_time(),
    pt.TxnField.last_valid: lambda txn: txn.last_valid(),
    pt.TxnField.note: lambda txn: txn.note(),
    pt.TxnField.lease: lambda txn: txn.lease(),
    pt.TxnField.receiver: lambda txn: txn.receiver(),
    pt.TxnField.amount: lambda txn: txn.amount(),
    pt.TxnField.close_remainder_to: lambda txn: txn.close_remainder_to(),
    pt.TxnField.vote_pk: lambda txn: txn.vote_pk(),
    pt.TxnField.selection_pk: lambda txn: txn.selection_pk(),
    pt.TxnField.vote_first: lambda txn: txn.vote_first(),
    pt.TxnField.vote_last: lambda txn: txn.vote_last(),
    pt.TxnField.vote_key_dilution: lambda txn: txn.vote_key_dilution(),
    pt.TxnField.type: lambda txn: txn.type(),
    pt.TxnField.type_enum: lambda txn: txn.type_enum(),
    pt.TxnField.xfer_asset: lambda txn: txn.xfer_asset(),
    pt.TxnField.asset_amount: lambda txn: txn.asset_amount(),
    pt.TxnField.asset_sender: lambda txn: txn.asset_sender(),
    pt.TxnField.asset_receiver: lambda txn: txn.asset_receiver(),
    pt.TxnField.asset_close_to: lambda txn: txn.asset_close_to(),
    pt.TxnField.group_index: lambda txn: txn.group_index(),
    pt.TxnField.tx_id: lambda txn: txn.tx_id(),
    pt.TxnField.application_id: lambda txn: txn.application_id(),
    pt.TxnField.on_completion: lambda txn: txn.on_completion(),
    pt.TxnField.approval_program: lambda txn: txn.approval_program(),
    pt.TxnField.clear_state_program: lambda txn: txn.clear_state_program(),
    pt.TxnField.rekey_to: lambda txn: txn.rekey_to(),
    pt.TxnField.config_asset: lambda txn: txn.config_asset(),
    pt.TxnField.config_asset_total: lambda txn: txn.config_asset_total(),
    pt.TxnField.config_asset_decimals: lambda txn: txn.config_asset_decimals(),
    pt.TxnField.config_asset_default_frozen: lambda txn: txn.config_asset_default_frozen(),
    pt.TxnField.config_asset_unit_name: lambda txn: txn.config_asset_unit_name(),
    pt.TxnField.config_asset_name: lambda txn: txn.config_asset_name(),
    pt.TxnField.config_asset_url: lambda txn: txn.config_asset_url(),
    pt.TxnField.config_asset_metadata_hash: lambda txn: txn.config_asset_metadata_hash(),
    pt.TxnField.config_asset_manager: lambda txn: txn.config_asset_manager(),
    pt.TxnField.config_asset_reserve: lambda txn: txn.config_asset_reserve(),
    pt.TxnField.config_asset_freeze: lambda txn: txn.config_asset_freeze(),
    pt.TxnField.config_asset_clawback: lambda txn: txn.config_asset_clawback(),
    pt.TxnField.freeze_asset: lambda txn: txn.freeze_asset(),
    pt.TxnField.freeze_asset_account: lambda txn: txn.freeze_asset_account(),
    pt.TxnField.freeze_asset_frozen: lambda txn: txn.freeze_asset_frozen(),
    pt.TxnField.global_num_uints: lambda txn: txn.global_num_uints(),
    pt.TxnField.global_num_byte_slices: lambda txn: txn.global_num_byte_slices(),
    pt.TxnField.local_num_uints: lambda txn: txn.local_num_uints(),
    pt.TxnField.local_num_byte_slices: lambda txn: txn.local_num_byte_slices(),
    pt.TxnField.extra_program_pages: lambda txn: txn.extra_program_pages(),
    pt.TxnField.nonparticipation: lambda txn: txn.nonparticipation(),
    pt.TxnField.created_asset_id: lambda txn: txn.created_asset_id(),
    pt.TxnField.created_application_id: lambda txn: txn.created_application_id(),
    pt.TxnField.last_log: lambda txn: txn.last_log(),
    pt.TxnField.state_proof_pk: lambda txn: txn.state_proof_pk(),
}

arrayFieldToProperty: Dict[pt.TxnField, Callable[[pt.TxnObject], pt.TxnArray]] = {
    pt.TxnField.application_args: lambda txn: txn.application_args,
    pt.TxnField.accounts: lambda txn: txn.accounts,
    pt.TxnField.assets: lambda txn: txn.assets,
    pt.TxnField.applications: lambda txn: txn.applications,
    pt.TxnField.logs: lambda txn: txn.logs,
    pt.TxnField.approval_program_pages: lambda txn: txn.approval_program_pages,
    pt.TxnField.clear_state_program_pages: lambda txn: txn.clear_state_program_pages,
}

arrayFieldToLengthField: Dict[pt.TxnField, pt.TxnField] = {
    pt.TxnField.application_args: pt.TxnField.num_app_args,
    pt.TxnField.accounts: pt.TxnField.num_accounts,
    pt.TxnField.assets: pt.TxnField.num_assets,
    pt.TxnField.applications: pt.TxnField.num_applications,
    pt.TxnField.logs: pt.TxnField.num_logs,
    pt.TxnField.approval_program_pages: pt.TxnField.num_approval_program_pages,
    pt.TxnField.clear_state_program_pages: pt.TxnField.num_clear_state_program_pages,
}


def test_txn_fields():
    dynamicGtxnArg = pt.Int(0)

    txnObjects = [
        (pt.Txn, pt.Op.txn, pt.Op.txna, pt.Op.txnas, [], []),
        *[
            (pt.Gtxn[i], pt.Op.gtxn, pt.Op.gtxna, pt.Op.gtxnas, [i], [])
            for i in range(pt.MAX_GROUP_SIZE)
        ],
        (
            pt.Gtxn[dynamicGtxnArg],
            pt.Op.gtxns,
            pt.Op.gtxnsa,
            pt.Op.gtxnsas,
            [],
            [pt.TealOp(dynamicGtxnArg, pt.Op.int, 0)],
        ),
        (pt.InnerTxn, pt.Op.itxn, pt.Op.itxna, pt.Op.itxnas, [], []),
        *[
            (pt.Gitxn[i], pt.Op.gitxn, pt.Op.gitxna, pt.Op.gitxnas, [i], [])
            for i in range(pt.MAX_GROUP_SIZE)
        ],
    ]

    for (
        txnObject,
        op,
        staticArrayOp,
        dynamicArrayOp,
        immediateArgsPrefix,
        irPrefix,
    ) in txnObjects:
        for field in pt.TxnField:
            if field.is_array:
                array = arrayFieldToProperty[field](txnObject)
                lengthExpr = array.length()

                lengthFieldName = arrayFieldToLengthField[field].arg_name
                immediateArgs = immediateArgsPrefix + [lengthFieldName]
                expected = pt.TealSimpleBlock(
                    irPrefix + [pt.TealOp(lengthExpr, op, *immediateArgs)]
                )
                expected.addIncoming()
                expected = pt.TealBlock.NormalizeBlocks(expected)

                version = max(op.min_version, field.min_version)

                actual, _ = lengthExpr.__teal__(pt.CompileOptions(version=version))
                actual.addIncoming()
                actual = pt.TealBlock.NormalizeBlocks(actual)

                assert (
                    actual == expected
                ), "{}: array length for field {} does not match expected".format(
                    op, field
                )

                if version > 2:
                    with pytest.raises(pt.TealInputError):
                        lengthExpr.__teal__(pt.CompileOptions(version=version - 1))

                for i in range(32):  # just an arbitrary large int
                    elementExpr = array[i]

                    immediateArgs = immediateArgsPrefix + [field.arg_name, i]
                    expected = pt.TealSimpleBlock(
                        irPrefix
                        + [pt.TealOp(elementExpr, staticArrayOp, *immediateArgs)]
                    )
                    expected.addIncoming()
                    expected = pt.TealBlock.NormalizeBlocks(expected)

                    version = max(staticArrayOp.min_version, field.min_version)

                    actual, _ = elementExpr.__teal__(pt.CompileOptions(version=version))
                    actual.addIncoming()
                    actual = pt.TealBlock.NormalizeBlocks(actual)

                    assert (
                        actual == expected
                    ), "{}: static array field {} does not match expected".format(
                        staticArrayOp, field
                    )

                    if version > 2:
                        with pytest.raises(pt.TealInputError):
                            elementExpr.__teal__(pt.CompileOptions(version=version - 1))

                if dynamicArrayOp is not None:
                    dynamicIndex = pt.Int(2)
                    dynamicElementExpr = array[dynamicIndex]

                    immediateArgs = immediateArgsPrefix + [field.arg_name]
                    expected = pt.TealSimpleBlock(
                        irPrefix
                        + [
                            pt.TealOp(dynamicIndex, pt.Op.int, 2),
                            pt.TealOp(
                                dynamicElementExpr, dynamicArrayOp, *immediateArgs
                            ),
                        ]
                    )
                    expected.addIncoming()
                    expected = pt.TealBlock.NormalizeBlocks(expected)

                    version = max(dynamicArrayOp.min_version, field.min_version)

                    actual, _ = dynamicElementExpr.__teal__(
                        pt.CompileOptions(version=version)
                    )
                    actual.addIncoming()
                    actual = pt.TealBlock.NormalizeBlocks(actual)

                    assert (
                        actual == expected
                    ), "{}: dynamic array field {} does not match expected".format(
                        dynamicArrayOp, field
                    )

                    if version > 2:
                        with pytest.raises(pt.TealInputError):
                            dynamicElementExpr.__teal__(
                                pt.CompileOptions(version=version - 1)
                            )

                continue

            if field in arrayFieldToLengthField.values():
                # ignore length fields since they are checked with their arrays
                continue

            if field == pt.TxnField.first_valid_time:
                # ignore first_valid_time since it is not exposed on pt.TxnObject yet
                continue

            expr = fieldToMethod[field](txnObject)

            immediateArgs = immediateArgsPrefix + [field.arg_name]
            expected = pt.TealSimpleBlock(
                irPrefix + [pt.TealOp(expr, op, *immediateArgs)]
            )
            expected.addIncoming()
            expected = pt.TealBlock.NormalizeBlocks(expected)

            version = max(op.min_version, field.min_version)

            actual, _ = expr.__teal__(pt.CompileOptions(version=version))
            actual.addIncoming()
            actual = pt.TealBlock.NormalizeBlocks(actual)

            assert actual == expected, "{}: field {} does not match expected".format(
                op, field
            )

            if version > 2:
                with pytest.raises(pt.TealInputError):
                    expr.__teal__(pt.CompileOptions(version=version - 1))



from itertools import product
from pathlib import Path
from typing import Any, Dict

import pytest

import pyteal as pt

from tests.compile_asserts import assert_teal_as_expected
from tests.blackbox import (
    Blackbox,
    BlackboxWrapper,
    algod_with_assertion,
    mode_to_execution_mode,
    PyTealDryRunExecutor,
)

from graviton.blackbox import (
    DryRunProperty as DRProp,
    DryRunEncoder as Encoder,
    DryRunExecutor,
    DryRunInspector,
    mode_has_property,
)

from graviton.invariant import Invariant

PATH = Path.cwd() / "tests" / "integration"
FIXTURES = PATH / "teal"
GENERATED = PATH / "generated"

# TODO: remove these skips after the following issue has been fixed https://github.com/algorand/pyteal/issues/199
STABLE_SLOT_GENERATION = False
SKIP_SCRATCH_ASSERTIONS = not STABLE_SLOT_GENERATION

# ---- Helper ---- #


def wrap_compile_and_save(
    subr, mode, version, assemble_constants, test_name, case_name
):
    is_app = mode == pt.Mode.Application

    teal = PyTealDryRunExecutor(subr, mode).compile(version, assemble_constants)
    tealfile = f'{"app" if is_app else "lsig"}_{case_name}.teal'

    tealdir = GENERATED / test_name
    tealdir.mkdir(parents=True, exist_ok=True)
    tealpath = tealdir / tealfile
    with open(tealpath, "w") as f:
        f.write(teal)

    print(
        f"""Subroutine {case_name}@{mode} generated TEAL.
saved to {tealpath}:
-------
{teal}
-------"""
    )

    return teal, is_app, tealfile


# ---- Subroutines for Blackbox Testing ---- #


@Blackbox(input_types=[])
@pt.Subroutine(pt.TealType.uint64)
def exp():
    return pt.Int(2) ** pt.Int(10)


@Blackbox(input_types=[pt.TealType.uint64])
@pt.Subroutine(pt.TealType.none)
def square_byref(x: pt.ScratchVar):
    return x.store(x.load() * x.load())


@Blackbox(input_types=[pt.TealType.uint64])
@pt.Subroutine(pt.TealType.uint64)
def square(x):
    return x ** pt.Int(2)


@Blackbox(input_types=[pt.TealType.anytype, pt.TealType.anytype])
@pt.Subroutine(pt.TealType.none)
def swap(x: pt.ScratchVar, y: pt.ScratchVar):
    z = pt.ScratchVar(pt.TealType.anytype)
    return pt.Seq(
        z.store(x.load()),
        x.store(y.load()),
        y.store(z.load()),
    )


@Blackbox(input_types=[pt.TealType.bytes, pt.TealType.uint64])
@pt.Subroutine(pt.TealType.bytes)
def string_mult(s: pt.ScratchVar, n):
    i = pt.ScratchVar(pt.TealType.uint64)
    tmp = pt.ScratchVar(pt.TealType.bytes)
    start = pt.Seq(i.store(pt.Int(1)), tmp.store(s.load()), s.store(pt.Bytes("")))
    step = i.store(i.load() + pt.Int(1))
    return pt.Seq(
        pt.For(start, i.load() <= n, step).Do(s.store(pt.Concat(s.load(), tmp.load()))),
        s.load(),
    )


@Blackbox(input_types=[pt.TealType.uint64])
@pt.Subroutine(pt.TealType.uint64)
def oldfac(n):
    return pt.If(n < pt.Int(2)).Then(pt.Int(1)).Else(n * oldfac(n - pt.Int(1)))


@Blackbox(input_types=[pt.TealType.uint64])
@pt.Subroutine(pt.TealType.uint64)
def slow_fibonacci(n):
    return (
        pt.If(n <= pt.Int(1))
        .Then(n)
        .Else(slow_fibonacci(n - pt.Int(2)) + slow_fibonacci(n - pt.Int(1)))
    )


def fac_with_overflow(n):
    if n < 2:
        return 1
    if n > 20:
        return 2432902008176640000
    return n * fac_with_overflow(n - 1)


def fib(n):
    a, b = 0, 1
    for _ in range(n):
        a, b = b, a + b
    return a


def fib_cost(args):
    cost = 17
    for n in range(1, args[0] + 1):
        cost += 31 * fib(n - 1)
    return cost


# ---- Blackbox pure unit tests (Skipping for now due to flakiness) ---- #


@pytest.mark.skipif(not STABLE_SLOT_GENERATION, reason="cf. #199")
@pytest.mark.parametrize(
    "subr, mode",
    product(
        [exp, square_byref, square, swap, string_mult, oldfac, slow_fibonacci],
        [pt.Mode.Application, pt.Mode.Signature],
    ),
)
def test_stable_teal_generation(subr, mode):
    """
    TODO: here's an example of issue #199 at play - need to run a dynamic version of `git bisect`
    to figure out what is driving this
    """
    case_name = subr.name()
    print(f"stable TEAL generation test for {case_name} in mode {mode}")

    _, _, tealfile = wrap_compile_and_save(subr, mode, 6, True, "stability", case_name)
    path2actual = GENERATED / "stability" / tealfile
    path2expected = FIXTURES / "stability" / tealfile
    assert_teal_as_expected(path2actual, path2expected)


APP_SCENARIOS = {
    exp: {
        "inputs": [()],
        # since only a single input, just assert a constant in each case
        "assertions": {
            DRProp.cost: 11,
            # int assertions on log outputs need encoding to varuint-hex:
            DRProp.lastLog: Encoder.hex(2**10),
            # dicts have a special meaning as assertions. So in the case of "finalScratch"
            # which is supposed to _ALSO_ output a dict, we need to use a lambda as a work-around
            DRProp.finalScratch: lambda _: {0: 1024},
            DRProp.stackTop: 1024,
            DRProp.maxStackHeight: 2,
            DRProp.status: "PASS",
            DRProp.passed: True,
            DRProp.rejected: False,
            DRProp.errorMessage: None,
        },
    },
    square_byref: {
        "inputs": [(i,) for i in range(100)],
        "assertions": {
            DRProp.cost: lambda _, actual: 20 < actual < 22,
            DRProp.lastLog: Encoder.hex(1337),
            # due to dry-run artifact of not reporting 0-valued scratchvars,
            # we have a special case for n=0:
            DRProp.finalScratch: lambda args, actual: (
                {1, 1337, (args[0] ** 2 if args[0] else 1)}
            ).issubset(set(actual.values())),
            DRProp.stackTop: 1337,
            DRProp.maxStackHeight: 3,
            DRProp.status: "PASS",
            DRProp.passed: True,
            DRProp.rejected: False,
            DRProp.errorMessage: None,
        },
    },
    square: {
        "inputs": [(i,) for i in range(100)],
        "assertions": {
            DRProp.cost: 14,
            DRProp.lastLog: {
                # since execution REJECTS for 0, expect last log for this case to be None
                (i,): Encoder.hex(i * i) if i else None
                for i in range(100)
            },
            DRProp.finalScratch: lambda args: (
                {0: args[0] ** 2, 1: args[0]} if args[0] else {}
            ),
            DRProp.stackTop: lambda args: args[0] ** 2,
            DRProp.maxStackHeight: 2,
            DRProp.status: lambda i: "PASS" if i[0] > 0 else "REJECT",
            DRProp.passed: lambda i: i[0] > 0,
            DRProp.rejected: lambda i: i[0] == 0,
            DRProp.errorMessage: None,
        },
    },
    swap: {
        "inputs": [(1, 2), (1, "two"), ("one", 2), ("one", "two")],
        "assertions": {
            DRProp.cost: 27,
            DRProp.lastLog: Encoder.hex(1337),
            DRProp.finalScratch: lambda args: {
                0: 1337,
                1: Encoder.hex0x(args[1]),
                2: Encoder.hex0x(args[0]),
                3: 1,
                4: 2,
                5: Encoder.hex0x(args[0]),
            },
            DRProp.stackTop: 1337,
            DRProp.maxStackHeight: 2,
            DRProp.status: "PASS",
            DRProp.passed: True,
            DRProp.rejected: False,
            DRProp.errorMessage: None,
        },
    },
    string_mult: {
        "inputs": [("xyzw", i) for i in range(100)],
        "assertions": {
            DRProp.cost: lambda args: 30 + 15 * args[1],
            DRProp.lastLog: (
                lambda args: Encoder.hex(args[0] * args[1]) if args[1] else None
            ),
            # due to dryrun 0-scratchvar artifact, special case for i == 0:
            DRProp.finalScratch: lambda args: (
                {
                    0: Encoder.hex0x(args[0] * args[1]),
                    1: Encoder.hex0x(args[0] * args[1]),
                    2: 1,
                    3: args[1],
                    4: args[1] + 1,
                    5: Encoder.hex0x(args[0]),
                }
                if args[1]
                else {
                    2: 1,
                    4: args[1] + 1,
                    5: Encoder.hex0x(args[0]),
                }
            ),
            DRProp.stackTop: lambda args: len(args[0] * args[1]),
            DRProp.maxStackHeight: lambda args: 3 if args[1] else 2,
            DRProp.status: lambda args: ("PASS" if 0 < args[1] < 45 else "REJECT"),
            DRProp.passed: lambda args: 0 < args[1] < 45,
            DRProp.rejected: lambda args: 0 >= args[1] or args[1] >= 45,
            DRProp.errorMessage: None,
        },
    },
    oldfac: {
        "inputs": [(i,) for i in range(25)],
        "assertions": {
            DRProp.cost: lambda args, actual: (
                actual - 40 <= 17 * args[0] <= actual + 40
            ),
            DRProp.lastLog: lambda args: (
                Encoder.hex(fac_with_overflow(args[0])) if args[0] < 21 else None
            ),
            DRProp.finalScratch: lambda args: (
                {1: args[0], 0: fac_with_overflow(args[0])}
                if 0 < args[0] < 21
                else (
                    {1: min(21, args[0])}
                    if args[0]
                    else {0: fac_with_overflow(args[0])}
                )
            ),
            DRProp.stackTop: lambda args: fac_with_overflow(args[0]),
            DRProp.maxStackHeight: lambda args: max(2, 2 * args[0]),
            DRProp.status: lambda args: "PASS" if args[0] < 21 else "REJECT",
            DRProp.passed: lambda args: args[0] < 21,
            DRProp.rejected: lambda args: args[0] >= 21,
            DRProp.errorMessage: lambda args, actual: (
                actual is None if args[0] < 21 else "overflowed" in actual
            ),
        },
    },
    slow_fibonacci: {
        "inputs": [(i,) for i in range(18)],
        "assertions": {
            DRProp.cost: lambda args: (fib_cost(args) if args[0] < 17 else 70_000),
            DRProp.lastLog: lambda args: (
                Encoder.hex(fib(args[0])) if 0 < args[0] < 17 else None
            ),
            DRProp.finalScratch: lambda args, actual: (
                actual == {1: args[0], 0: fib(args[0])}
                if 0 < args[0] < 17
                else (True if args[0] >= 17 else actual == {})
            ),
            # we declare to "not care" about the top of the stack for n >= 17
            DRProp.stackTop: lambda args, actual: (
                actual == fib(args[0]) if args[0] < 17 else True
            ),
            # similarly, we don't care about max stack height for n >= 17
            DRProp.maxStackHeight: lambda args, actual: (
                actual == max(2, 2 * args[0]) if args[0] < 17 else True
            ),
            DRProp.status: lambda args: "PASS" if 0 < args[0] < 8 else "REJECT",
            DRProp.passed: lambda args: 0 < args[0] < 8,
            DRProp.rejected: lambda args: 0 >= args[0] or args[0] >= 8,
            DRProp.errorMessage: lambda args, actual: (
                actual is None
                if args[0] < 17
                else "dynamic cost budget exceeded" in actual
            ),
        },
    },
}

# NOTE: logic sig dry runs are missing some information when compared with app dry runs.
# Therefore, certain assertions don't make sense for logic sigs explaining why some of the below are commented out:
LOGICSIG_SCENARIOS = {
    exp: {
        "inputs": [()],
        "assertions": {
            # DRProp.cost: 11,
            # DRProp.lastLog: Encoder.hex(2 ** 10),
            DRProp.finalScratch: lambda _: {},
            DRProp.stackTop: 1024,
            DRProp.maxStackHeight: 2,
            DRProp.status: "PASS",
            DRProp.passed: True,
            DRProp.rejected: False,
            DRProp.errorMessage: None,
        },
    },
    square_byref: {
        "inputs": [(i,) for i in range(100)],
        "assertions": {
            # DRProp.cost: lambda _, actual: 20 < actual < 22,
            # DRProp.lastLog: Encoder.hex(1337),
            # due to dry-run artifact of not reporting 0-valued scratchvars,
            # we have a special case for n=0:
            DRProp.finalScratch: lambda args: (
                {0: 1, 1: args[0] ** 2} if args[0] else {0: 1}
            ),
            DRProp.stackTop: 1337,
            DRProp.maxStackHeight: 3,
            DRProp.status: "PASS",
            DRProp.passed: True,
            DRProp.rejected: False,
            DRProp.errorMessage: None,
        },
    },
    square: {
        "inputs": [(i,) for i in range(100)],
        "assertions": {
            # DRProp.cost: 14,
            # DRProp.lastLog: {(i,): Encoder.hex(i * i) if i else None for i in range(100)},
            DRProp.finalScratch: lambda args: ({0: args[0]} if args[0] else {}),
            DRProp.stackTop: lambda args: args[0] ** 2,
            DRProp.maxStackHeight: 2,
            DRProp.status: lambda i: "PASS" if i[0] > 0 else "REJECT",
            DRProp.passed: lambda i: i[0] > 0,
            DRProp.rejected: lambda i: i[0] == 0,
            DRProp.errorMessage: None,
        },
    },
    swap: {
        "inputs": [(1, 2), (1, "two"), ("one", 2), ("one", "two")],
        "assertions": {
            # DRProp.cost: 27,
            # DRProp.lastLog: Encoder.hex(1337),
            DRProp.finalScratch: lambda args: {
                0: 3,
                1: 4,
                2: Encoder.hex0x(args[0]),
                3: Encoder.hex0x(args[1]),
                4: Encoder.hex0x(args[0]),
            },
            DRProp.stackTop: 1337,
            DRProp.maxStackHeight: 2,
            DRProp.status: "PASS",
            DRProp.passed: True,
            DRProp.rejected: False,
            DRProp.errorMessage: None,
        },
    },
    string_mult: {
        "inputs": [("xyzw", i) for i in range(100)],
        "assertions": {
            # DRProp.cost: lambda args: 30 + 15 * args[1],
            # DRProp.lastLog: lambda args: Encoder.hex(args[0] * args[1]) if args[1] else None,
            DRProp.finalScratch: lambda args: (
                {
                    0: len(args[0]),
                    1: args[1],
                    2: args[1] + 1,
                    3: Encoder.hex0x(args[0]),
                    4: Encoder.hex0x(args[0] * args[1]),
                }
                if args[1]
                else {
                    0: len(args[0]),
                    2: args[1] + 1,
                    3: Encoder.hex0x(args[0]),
                }
            ),
            DRProp.stackTop: lambda args: len(args[0] * args[1]),
            DRProp.maxStackHeight: lambda args: 3 if args[1] else 2,
            DRProp.status: lambda args: "PASS" if args[1] else "REJECT",
            DRProp.passed: lambda args: bool(args[1]),
            DRProp.rejected: lambda args: not bool(args[1]),
            DRProp.errorMessage: None,
        },
    },
    oldfac: {
        "inputs": [(i,) for i in range(25)],
        "assertions": {
            # DRProp.cost: lambda args, actual: actual - 40 <= 17 * args[0] <= actual + 40,
            # DRProp.lastLog: lambda args, actual: (actual is None) or (int(actual, base=16) == fac_with_overflow(args[0])),
            DRProp.finalScratch: lambda args: (
                {0: min(args[0], 21)} if args[0] else {}
            ),
            DRProp.stackTop: lambda args: fac_with_overflow(args[0]),
            DRProp.maxStackHeight: lambda args: max(2, 2 * args[0]),
            DRProp.status: lambda args: "PASS" if args[0] < 21 else "REJECT",
            DRProp.passed: lambda args: args[0] < 21,
            DRProp.rejected: lambda args: args[0] >= 21,
            DRProp.errorMessage: lambda args, actual: (
                actual is None
                if args[0] < 21
                else "logic 0 failed at line 21: * overflowed" in actual
            ),
        },
    },
    slow_fibonacci: {
        "inputs": [(i,) for i in range(18)],
        "assertions": {
            # DRProp.cost: fib_cost,
            # DRProp.lastLog: fib_last_log,
            # by returning True for n >= 15, we're declaring that we don't care about the scratchvar's for such cases:
            DRProp.finalScratch: lambda args, actual: (
                actual == {0: args[0]}
                if 0 < args[0] < 15
                else (True if args[0] else actual == {})
            ),
            DRProp.stackTop: lambda args, actual: (
                actual == fib(args[0]) if args[0] < 15 else True
            ),
            DRProp.maxStackHeight: lambda args, actual: (
                actual == max(2, 2 * args[0]) if args[0] < 15 else True
            ),
            DRProp.status: lambda args: "PASS" if 0 < args[0] < 15 else "REJECT",
            DRProp.passed: lambda args: 0 < args[0] < 15,
            DRProp.rejected: lambda args: not (0 < args[0] < 15),
            DRProp.errorMessage: lambda args, actual: (
                actual is None
                if args[0] < 15
                else "dynamic cost budget exceeded" in actual
            ),
        },
    },
}


def blackbox_test_runner(
    subr: pt.SubroutineFnWrapper,
    mode: pt.Mode,
    scenario: Dict[str, Any],
    version: int,
    assemble_constants: bool = True,
):
    case_name = subr.name()
    print(f"blackbox test of {case_name} with mode {mode}")
    exec_mode = mode_to_execution_mode(mode)

    # 0. Validations
    assert isinstance(subr, BlackboxWrapper), f"unexpected subr type {type(subr)}"
    assert isinstance(mode, pt.Mode)

    # 1. Compile to TEAL
    teal, _, tealfile = wrap_compile_and_save(
        subr, mode, version, assemble_constants, "blackbox", case_name
    )

    # Fail fast in case algod is not configured:
    algod = algod_with_assertion()

    # 2. validate dry run scenarios:
    inputs, predicates = Invariant.inputs_and_invariants(
        scenario, exec_mode, raw_predicates=True
    )

    # 3. execute dry run sequence:
    execute = DryRunExecutor.execute_one_dryrun
    inspectors = list(map(lambda a: execute(algod, teal, a, exec_mode), inputs))

    # 4. Statistical report:
    csvpath = GENERATED / "blackbox" / f"{tealfile}.csv"
    with open(csvpath, "w") as f:
        f.write(DryRunInspector.csv_report(inputs, inspectors))

    print(f"Saved Dry Run CSV report to {csvpath}")

    # 5. Sequential assertions (if provided any)
    for i, type_n_assertion in enumerate(predicates.items()):
        dr_prop, predicate = type_n_assertion

        if SKIP_SCRATCH_ASSERTIONS and dr_prop == DRProp.finalScratch:
            print("skipping scratch assertions because unstable slots produced")
            continue

        assert mode_has_property(exec_mode, dr_prop)

        invariant = Invariant(predicate, name=f"{case_name}[{i}]@{mode}-{dr_prop}")
        print(f"{i+1}. Assertion for {case_name}-{mode}: {dr_prop} <<{predicate}>>")
        invariant.validates(dr_prop, inputs, inspectors)


# ---- Graviton / Blackbox tests ---- #


@pytest.mark.parametrize("subr, scenario", APP_SCENARIOS.items())
def test_blackbox_subroutines_as_apps(
    subr: pt.SubroutineFnWrapper,
    scenario: Dict[str, Any],
):
    blackbox_test_runner(subr, pt.Mode.Application, scenario, 6)


@pytest.mark.parametrize("subr, scenario", LOGICSIG_SCENARIOS.items())
def test_blackbox_subroutines_as_logic_sigs(
    subr: pt.SubroutineFnWrapper,
    scenario: Dict[str, Any],
):
    blackbox_test_runner(subr, pt.Mode.Signature, scenario, 6)


def blackbox_pyteal_example1():
    # Example 1: Using blackbox_pyteal for a simple test of both an app and logic sig:
    from graviton.blackbox import DryRunEncoder

    from pyteal import Int, Mode, Subroutine, TealType
    from tests.blackbox import Blackbox

    @Blackbox(input_types=[TealType.uint64])
    @Subroutine(TealType.uint64)
    def square(x):
        return x ** Int(2)

    # provide args for evaluation (will compute x^2)
    x = 9
    args = [x]

    # evaluate the programs
    app_result = PyTealDryRunExecutor(square, Mode.Application).dryrun(args)
    lsig_result = PyTealDryRunExecutor(square, Mode.Signature).dryrun(args)

    # check to see that x^2 is at the top of the stack as expected
    assert app_result.stack_top() == x**2, app_result.report(
        args, "stack_top() gave unexpected results for app"
    )
    assert lsig_result.stack_top() == x**2, lsig_result.report(
        args, "stack_top() gave unexpected results for lsig"
    )

    # check to see that itob of x^2 has been logged (only for the app case)
    assert app_result.last_log() == DryRunEncoder.hex(x**2), app_result.report(
        args, "last_log() gave unexpected results from app"
    )


def blackbox_pyteal_example2():
    # Example 2: Using blackbox_pyteal to make 400 assertions and generate a CSV report with 400 dryrun rows
    from itertools import product
    import math
    from pathlib import Path
    import random

    from graviton.blackbox import DryRunInspector

    from pyteal import (
        For,
        If,
        Int,
        Mod,
        Mode,
        ScratchVar,
        Seq,
        Subroutine,
        TealType,
    )

    from tests.blackbox import Blackbox

    # GCD via the Euclidean Algorithm (iterative version):
    @Blackbox(input_types=[TealType.uint64, TealType.uint64])
    @Subroutine(TealType.uint64)
    def euclid(x, y):
        a = ScratchVar(TealType.uint64)
        b = ScratchVar(TealType.uint64)
        tmp = ScratchVar(TealType.uint64)
        start = If(x < y, Seq(a.store(y), b.store(x)), Seq(a.store(x), b.store(y)))
        cond = b.load() > Int(0)
        step = Seq(
            tmp.store(b.load()), b.store(Mod(a.load(), b.load())), a.store(tmp.load())
        )
        return Seq(For(start, cond, step).Do(Seq()), a.load())

    # generate a report with 400 = 20*20 dry run rows:
    N = 20
    inputs = list(
        product(
            tuple(random.randint(0, 1000) for _ in range(N)),
            tuple(random.randint(0, 1000) for _ in range(N)),
        )
    )

    # assert that each result is that same as what Python's math.gcd() computes
    inspectors = PyTealDryRunExecutor(euclid, Mode.Application).dryrun_on_sequence(
        inputs
    )
    for i, result in enumerate(inspectors):
        args = inputs[i]
        assert result.stack_top() == math.gcd(*args), result.report(
            args, f"failed for {args}"
        )

    # save the CSV to ...current working directory.../euclid.csv
    euclid_csv = DryRunInspector.csv_report(inputs, inspectors)
    with open(Path.cwd() / "euclid.csv", "w") as f:
        f.write(euclid_csv)


def blackbox_pyteal_example3():
    # Example 3: declarative Test Driven Development approach through Invariant's
    from itertools import product
    import math
    import random

    from graviton.blackbox import (
        DryRunEncoder,
        DryRunProperty as DRProp,
    )
    from graviton.invariant import Invariant

    from pyteal import If, Int, Mod, Mode, Subroutine, TealType

    from tests.blackbox import Blackbox

    # avoid flaky tests just in case I was wrong about the stack height invariant...
    random.seed(42)

    # helper that will be used for scratch-slots invariant:
    def is_subdict(x, y):
        return all(k in y and x[k] == y[k] for k in x)

    predicates = {
        # the program's log should be the hex encoding of Python's math.gcd:
        DRProp.lastLog: lambda args: (
            DryRunEncoder.hex(math.gcd(*args)) if math.gcd(*args) else None
        ),
        # the program's scratch should contain math.gcd() at slot 0:
        DRProp.finalScratch: lambda args, actual: is_subdict(
            {0: math.gcd(*args)}, actual
        ),
        # the top of the stack should be math.gcd():
        DRProp.stackTop: lambda args: math.gcd(*args),
        # Making the rather weak assertion that the max stack height is between 2 and 3*log2(max(args)):
        DRProp.maxStackHeight: (
            lambda args, actual: 2
            <= actual
            <= 3 * math.ceil(math.log2(max(args + (1,))))
        ),
        # the program PASS'es exactly for non-0 math.gcd (3 variants):
        DRProp.status: lambda args: "PASS" if math.gcd(*args) else "REJECT",
        DRProp.passed: lambda args: bool(math.gcd(*args)),
        DRProp.rejected: lambda args: not bool(math.gcd(*args)),
        # the program never errors:
        DRProp.errorMessage: None,
    }

    # Define a scenario 400 random pairs (x,y) as inputs:
    N = 20
    inputs = list(
        product(
            tuple(random.randint(0, 1000) for _ in range(N)),
            tuple(random.randint(0, 1000) for _ in range(N)),
        )
    )

    # GCD via the Euclidean Algorithm (recursive version):
    @Blackbox(input_types=[TealType.uint64, TealType.uint64])
    @Subroutine(TealType.uint64)
    def euclid(x, y):
        return (
            If(x < y)
            .Then(euclid(y, x))
            .Else(If(y == Int(0)).Then(x).Else(euclid(y, Mod(x, y))))
        )

    # Execute on the input sequence to get a dry-run inspectors:
    inspectors = PyTealDryRunExecutor(euclid, Mode.Application).dryrun_on_sequence(
        inputs
    )

    # Assert that each invariant holds on the sequences of inputs and dry-runs:
    for property, predicate in predicates.items():
        Invariant(predicate).validates(property, inputs, inspectors)


def blackbox_pyteal_example4():
    # Example 4: Using PyTealDryRunExecutor to debug an ABIReturnSubroutine with an app, logic sig and csv report
    from pathlib import Path
    import random

    from graviton.blackbox import DryRunInspector

    from pyteal import (
        abi,
        ABIReturnSubroutine,
        Expr,
        For,
        Int,
        Mode,
        ScratchVar,
        Seq,
        TealType,
    )

    from tests.blackbox import Blackbox, PyTealDryRunExecutor

    # Sum a dynamic uint64 array
    @Blackbox(input_types=[None])
    @ABIReturnSubroutine
    def abi_sum(toSum: abi.DynamicArray[abi.Uint64], *, output: abi.Uint64) -> Expr:
        i = ScratchVar(TealType.uint64)
        valueAtIndex = abi.Uint64()
        return Seq(
            output.set(0),
            For(
                i.store(Int(0)),
                i.load() < toSum.length(),
                i.store(i.load() + Int(1)),
            ).Do(
                Seq(
                    toSum[i.load()].store_into(valueAtIndex),
                    output.set(output.get() + valueAtIndex.get()),
                )
            ),
        )

    # instantiate PyTealDryRunExecutor objects for the app and lsig:
    app_pytealer = PyTealDryRunExecutor(abi_sum, Mode.Application)
    lsig_pytealer = PyTealDryRunExecutor(abi_sum, Mode.Signature)

    # generate reports with the same random inputs (fix the randomness with a seed):
    random.seed(42)

    N = 50  # the number of dry runs for each experiment
    choices = range(10_000)
    inputs = []
    for n in range(N):
        inputs.append(tuple([random.sample(choices, n)]))

    app_inspectors = app_pytealer.dryrun_on_sequence(inputs)

    lsig_inspectors = lsig_pytealer.dryrun_on_sequence(inputs)

    for i in range(N):
        args = inputs[i]

        app_inspector = app_inspectors[i]
        lsig_inspector = lsig_inspectors[i]

        def message(insp):
            return insp.report(args, f"failed for {args}", row=i)

        # the app should pass exactly when it's cost was within the 700 budget:
        assert app_inspector.passed() == (app_inspector.cost() <= 700), message(
            app_inspector
        )
        # the lsig always passes (never goes over budget):
        assert lsig_inspector.passed(), message(lsig_inspector)

        expected = sum(args[0])
        actual4app = app_inspector.last_log()
        assert expected == actual4app, message(app_inspector)

        if i > 0:
            assert expected in app_inspector.final_scratch().values(), message(
                app_inspector
            )
            assert expected in lsig_inspector.final_scratch().values(), message(
                lsig_inspector
            )

    def report(kind):
        assert kind in ("app", "lsig")
        insps = app_inspectors if kind == "app" else lsig_inspectors
        csv_report = DryRunInspector.csv_report(inputs, insps)
        with open(Path.cwd() / f"abi_sum_{kind}.csv", "w") as f:
            f.write(csv_report)

    report("app")
    report("lsig")


def blackbox_pyteal_example5():
    from graviton.blackbox import DryRunEncoder

    from pyteal import abi, Subroutine, TealType, Int, Mode
    from tests.blackbox import Blackbox

    @Blackbox([None])
    @Subroutine(TealType.uint64)
    def cubed(n: abi.Uint64):
        return n.get() ** Int(3)

    app_pytealer = PyTealDryRunExecutor(cubed, Mode.Application)
    lsig_pytealer = PyTealDryRunExecutor(cubed, Mode.Signature)

    inputs = [[i] for i in range(1, 11)]

    app_inspect = app_pytealer.dryrun_on_sequence(inputs)
    lsig_inspect = lsig_pytealer.dryrun_on_sequence(inputs)

    for index, inspect in enumerate(app_inspect):
        input_var = inputs[index][0]
        assert inspect.stack_top() == input_var**3, inspect.report(
            args=inputs[index], msg="stack_top() gave unexpected results from app"
        )
        assert inspect.last_log() == DryRunEncoder.hex(input_var**3), inspect.report(
            args=inputs[index], msg="last_log() gave unexpected results from app"
        )

    for index, inspect in enumerate(lsig_inspect):
        input_var = inputs[index][0]
        assert inspect.stack_top() == input_var**3, inspect.report(
            args=inputs[index], msg="stack_top() gave unexpected results from app"
        )


def blackbox_pyteal_while_continue_test():
    from tests.blackbox import Blackbox
    from pyteal import (
        Continue,
        Int,
        Mode,
        Return,
        ScratchVar,
        Seq,
        Subroutine,
        TealType,
        While,
    )

    @Blackbox(input_types=[TealType.uint64])
    @Subroutine(TealType.uint64)
    def while_continue_accumulation(n):
        i = ScratchVar(TealType.uint64)
        return Seq(
            i.store(Int(0)),
            While(i.load() < n).Do(
                Seq(
                    i.store(i.load() + Int(1)),
                    Continue(),
                )
            ),
            Return(i.load()),
        )

    for x in range(30):
        args = [x]
        lsig_result = PyTealDryRunExecutor(
            while_continue_accumulation, Mode.Signature
        ).dryrun(args)
        if x == 0:
            assert not lsig_result.passed()
        else:
            assert lsig_result.passed()

        assert lsig_result.stack_top() == x, lsig_result.report(
            args, "stack_top() gave unexpected results for lsig"
        )


def blackbox_pyteal_named_tupleness_test():
    from typing import Literal as L
    from tests.blackbox import Blackbox
    from pyteal import (
        Seq,
        abi,
        Subroutine,
        TealType,
        Return,
        And,
        Mode,
    )

    class NamedTupleExample(abi.NamedTuple):
        a: abi.Field[abi.Bool]
        b: abi.Field[abi.Address]
        c: abi.Field[abi.Tuple2[abi.Uint64, abi.Bool]]
        d: abi.Field[abi.StaticArray[abi.Byte, L[10]]]
        e: abi.Field[abi.StaticArray[abi.Bool, L[4]]]
        f: abi.Field[abi.Uint64]

    @Blackbox(input_types=[None] * 6)
    @Subroutine(TealType.uint64)
    def named_tuple_field_access(
        a_0: abi.Bool,
        a_1: abi.Address,
        a_2: abi.Tuple2[abi.Uint64, abi.Bool],
        a_3: abi.StaticArray[abi.Byte, L[10]],
        a_4: abi.StaticArray[abi.Bool, L[4]],
        a_5: abi.Uint64,
    ):
        return Seq(
            (v_tuple := NamedTupleExample()).set(a_0, a_1, a_2, a_3, a_4, a_5),
            (v_a := abi.Bool()).set(v_tuple.a),
            (v_b := abi.Address()).set(v_tuple.b),
            (v_c := abi.make(abi.Tuple2[abi.Uint64, abi.Bool])).set(v_tuple.c),
            (v_d := abi.make(abi.StaticArray[abi.Byte, L[10]])).set(v_tuple.d),
            (v_e := abi.make(abi.StaticArray[abi.Bool, L[4]])).set(v_tuple.e),
            (v_f := abi.Uint64()).set(v_tuple.f),
            Return(
                And(
                    a_0.get() == v_a.get(),
                    a_1.get() == v_b.get(),
                    a_2.encode() == v_c.encode(),
                    a_3.encode() == v_d.encode(),
                    a_4.encode() == v_e.encode(),
                    a_5.get() == v_f.get(),
                )
            ),
        )

    lsig_pytealer = PyTealDryRunExecutor(named_tuple_field_access, Mode.Signature)
    args = (False, b"1" * 32, (0, False), b"0" * 10, [True] * 4, 0)

    inspector = lsig_pytealer.dryrun(args)

    assert inspector.stack_top() == 1
    assert inspector.passed()


@pytest.mark.parametrize(
    "example",
    [
        blackbox_pyteal_example1,
        blackbox_pyteal_example2,
        blackbox_pyteal_example3,
        blackbox_pyteal_example4,
        blackbox_pyteal_example5,
        blackbox_pyteal_while_continue_test,
        blackbox_pyteal_named_tupleness_test,
    ],
)
def test_blackbox_pyteal_examples(example):
    example()
